if (window.location.protocol == 'https:') {
  window.location.href = 'http:' + window.location.href.substring(window.location.protocol.length);
} else if (window.location.hostname == 'autocode.crystal.sh') {
  window.location.href = 'http://app.autocode.run';
}

CodeMirror.registerHelper('hint', 'anyword', function(editor, options) {
  var WORD = /[\w$]+/, RANGE = 500;
  
  var word = options && options.word || WORD;
  var range = options && options.range || RANGE;
  var cur = editor.getCursor(), curLine = editor.getLine(cur.line);
  var end = cur.ch, start = end;
  while (start && word.test(curLine.charAt(start - 1))) --start;
  var curWord = start != end && curLine.slice(start, end);

  var list = [];
  for (var i in autocode.data.autocomplete) {
    if (curWord === false || autocode.data.autocomplete[i].match(curWord.replace(new RegExp('\\$', 'g'), '\\$'), 'i')) {
      var word = autocode.data.autocomplete[i];
      if (word.substr(0, 1) != '$') {
        word += ': ';
      }
      if (list.indexOf(autocode.data.autocomplete[i]) === -1) {
        list.push(word);
      }
    }
  }
  
  return {
    list: list,
    from: CodeMirror.Pos(cur.line, start),
    to: CodeMirror.Pos(cur.line, end)
  };
});

jQuery.fn.extend({
  visibleHeight: function() {
    var o = $(this);
    if (o.is(':hidden')) {
      return 0;
    } else {
      return o.outerHeight();
    }
  },
  visibleWidth: function() {
    var o = $(this);
    if (o.is(':hidden')) {
      return 0;
    } else {
      return o.outerWidth();
    }
  }
});

var autocode = {
  action: {},
  config: {},
  data: {
    current: {
      pin: true
    },
    exportTypes: ['engine','generator','helper','processor','schema','spec','transformer'],
    formats: [
      'c',
      'clojure',
      'cpp',
      'csharp',
      'cson',
      'coffeescript',
      'd',
      'dart',
      'dockerfile',
      'erlang',
      'fsharp',
      'go',
      'haskell',
      'haxe',
      'html',
      'ini',
      'java',
      'javascript',
      'json',
      'markdown',
      'objectivec',
      'perl',
      'php',
      'python',
      'r',
      'ruby',
      'rust',
      'scala',
      'text',
      'yaml',
      'xml'
    ]
  },
  state: {},
  user: {},
  initState: function() {
    $('a').each(function() {
      if ($(this).data('state')) {
        return;
      }
      
      $(this).data('state', true);
      
      if (navigator.userAgent.match(/mobile/i)) {
        $(this).bind({
          touchstart: function() {
            $(window).removeData('scrolled');
          },
          touchend: function(e) {
            if ($(window).data('scrolled')) {
              return false;
            }
            return autocode.initStateCallback(e, $(this));
          }
        });
      } else {
        $(this).click(autocode.initStateCallback);
      }
    });
  },
  initStateCallback: function(e, o) {
    var o = o || $(this);
    var href = o.attr('href');
    if (!href) {
      return false;
    } else if (href.match(/^https?:/)) {
      // open new window
      window.open(href, '_blank');
      
      // close popover
      autocode.popover.close();
      
      return false;
    }
    
    e.preventDefault();
    
    // get query
    var query = href.split('?');
    href = query[0];
    query = autocode.query.search(query[1]);
    
    autocode.ga.send('pageview', href);
    
    console.log(href);
    
    // get name
    var name = href.split('/').splice(0, 2).join('/');
    
    // get state name
    var state_name = href;
    if (name == autocode.repo) {
      state_name = state_name.split('/').splice(2).join('/');
    }
    var state = autocode.state[state_name];
    if (state) {
      $('input:focus').blur();
      state(query);
      autocode.initState();
    } else if (!state) {
      autocode.action.loadProject({ force: true, name: name });
    }
    
    history.pushState(null, null, href);
    
    if (autocode.listener.listeners[state_name]) {
      for (var listener_name in autocode.listener.listeners[state_name]) {
        autocode.listener.listeners[state_name][listener_name](query);
      }
    }
    
    autocode.resize.all();
    
    return false;
  }
};

$(window).load(autocode.init);
$(window).resize(autocode.resize);;autocode.api = {
  url: null,
  ajax: function(opts) {
    opts = opts || {};
    
    if (!opts.url) {
      opts.url = autocode.api.url;
    }
    opts.url += opts.uri;
    
    var ajax = {
      contentType: 'application/json',
      dataType: 'json',
      method: opts.method,
      url: opts.url,
      complete: opts.complete,
      error: function(data) {
        if (opts.error) {
          opts.error(data.responseJSON);
        }
      },
      success: opts.success,
      xhrFields: {
        withCredentials: true
      }
    };
    
    if (opts.data) {
      if (opts.method == 'get') {
        for (var name in opts.data) {
          if (ajax.url.match(/\?/)) {
            ajax.url += '&';
          } else {
            ajax.url += '?';
          }
          ajax.url += name + '=' + opts.data[name];
        }
      } else {
        ajax.data = JSON.stringify(opts.data);
      }
    }
    
    return $.ajax(ajax);
  },
  get: function(opts) {
    opts.method = 'get';
    return autocode.api.ajax(opts);
  },
  post: function(opts) {
    opts.method = 'post';
    return autocode.api.ajax(opts);
  }
};;autocode.console = {
  init: function() {
    var showConsole = autocode.storage.get('showConsole');
    if (showConsole) {
      $('#console').show();
    } else {
      $('#console').hide();
    }
  }
};;autocode.current = {
  fromExports: function(items, type) {
    if (!autocode.project.exports) {
      return {};
    }
    
    for (var export_name in autocode.project.exports) {
      Export = autocode.project.exports[export_name];
      if (Export.type == type) {
        items[export_name] = Export;
      }
    }
    
    return items;
  },
  fromImports: function(items, type) {
    if (!autocode.data[type]) {
      return {};
    }
    
    for (var import_name in autocode.data[type]) {
      items[import_name] = autocode.data[type][import_name];
    }
    
    return items;
  },
  generators: function() {
    var schemas = {};
    schemas = autocode.current.fromExports(schemas, 'generator');
    schemas = autocode.current.fromImports(schemas, 'generators');
    return schemas;
  },
  schemas: function() {
    var schemas = {};
    schemas = autocode.current.fromExports(schemas, 'schema');
    schemas = autocode.current.fromImports(schemas, 'schemas');
    return schemas;
  }
};;autocode.editor = {
  defaultEditor: 'codemirror',
  defaultTheme: 'dracula',
  codemirror: {
    themes: [
      '3024-day',
      '3024-night',
      'abcdef',
      'ambiance-mobile',
      'ambiance',
      'base16-dark',
      'base16-light',
      'blackboard',
      'cobalt',
      'colorforth',
      'dracula',
      'eclipse',
      'elegant',
      'erlang-dark',
      'icecoder',
      'lesser-dark',
      'liquibyte',
      'material',
      'mbo',
      'mdn-like',
      'midnight',
      'monokai',
      'neat',
      'neo',
      'night',
      'paraiso-dark',
      'paraiso-light',
      'pastel-on-dark',
      'rubyblue',
      'seti',
      'solarized',
      'the-matrix',
      'tomorrow-night-bright',
      'tomorrow-night-eighties',
      'ttcn',
      'twilight',
      'vibrant-ink',
      'xq-dark',
      'xq-light',
      'yeti',
      'zenburn'
    ]
  },
  enableLineNumbers: function() {
    autocode.storage.set('lineNumbers', true);
    $('.CodeMirror').each(function() {
      $(this)[0].CodeMirror.setOption('lineNumbers', true);
    });
  },
  disableLineNumbers: function() {
    autocode.storage.remove('lineNumbers');
    $('.CodeMirror').each(function() {
      $(this)[0].CodeMirror.setOption('lineNumbers', false);
    });
  },
  getTheme: function() {
    var theme = autocode.storage.get('theme');
    if (!theme) {
      theme = autocode.editor.defaultTheme;
    }
    return theme;
  },
  init: function() {
    var theme = autocode.editor.getTheme();
    $('head').append('<link id="' + theme + '-theme" href="components/CodeMirror/theme/' + theme + '.css" rel="stylesheet" type="text/css" />');
  },
  lineNumbersEnabled: function() {
    var lineNumbers = autocode.storage.get('lineNumbers');
    return lineNumbers ? true : false;
  },
  setTheme: function(theme) {
    autocode.storage.set('theme', theme);
    if (!$('#' + theme + '-theme').length) {
      $('head').append('<link id="' + theme + '-theme" href="components/CodeMirror/theme/' + theme + '.css" rel="stylesheet" type="text/css" />');
    }
    $('.CodeMirror').each(function() {
      $(this)[0].CodeMirror.setOption('theme', theme);
    });
  }
};;autocode.element = {};;autocode.file = {
  extensions: {
    bowerrc: 'json',
    coffee: 'coffeescript',
    dockerfile: 'docker',
    dockerignore: 'docker',
    eslintrc: 'eslint',
    gitignore: 'git',
    gitkeep: 'git',
    htm: 'html',
    html: 'html',
    js: 'js',
    jshintrc: 'json',
    json: 'json',
    md: 'markdown',
    npmignore: 'npm',
    yml: 'yaml',
    yaml: 'yaml'
  },
  icon: function(filename) {
    var extension = filename.split('.');
    if (extension.length == 1) {
      extension = extension[0];
    } else {
      extension = extension[extension.length-1];
    }
    extension = extension.toLowerCase();
    
    var icon = 'file';
    if (autocode.file.extensions[extension]) {
      icon = autocode.file.extensions[extension];
    }
    
    return icon;
  }
};;autocode.fuzzy = {
  close: function() {
    $('#fuzzy').remove();
  },
  open: function(opts) {
    var fuzzy = $('#fuzzy');
    if (!fuzzy.length) {
      fuzzy = $(document.createElement('div'));
      $('body').append(fuzzy);
    }
    fuzzy.attr('id', 'fuzzy');
    fuzzy.data('target', opts.target);
    
    var icons_exist = false;
    for (var i = 0; i < opts.rows.length; i++) {
      if (opts.rows[i].icon) {
        icons_exist = true;
        break;
      }
    }
    
    var table = $(document.createElement('div'));
    table.addClass('table');
    
    var row_icon,
      row_link,
      row_text;
    for (var i = 0; i < opts.rows.length; i++) {
      row = opts.rows[i];
      
      row_link = $(document.createElement('a'));
      if (row.state) {
        row_link.attr('href', row.state);
      } else if (typeof(row.action) == 'object') {
        row_link.attr('onclick', 'autocode.action[\'' + row.action.name + '\'](' + JSON.stringify(row.action.data) + ')');
      } else if (typeof(row.action) == 'string') {
        row_link.attr('onclick', 'autocode.action[\'' + row.action + '\']()');
      }
      table.append(row_link);
      
      if (icons_exist) {
        row_icon = $(document.createElement('span'));
        row_icon.addClass('icon');
        if (row.icon) {
          if (row.icon.match(/^https?:/)) {
            row_icon.css('background-image', 'url(' + row.icon + ')');
          } else {
            row_icon.addClass(row.icon);
          }
        }
        if (row.style == 'divider') {
          row_icon.css('border-top', '1px #CCC solid');
        }
        row_link.append(row_icon);
      }
      
      row_text = $(document.createElement('span'));
      row_text.addClass('text');
      row_text.html(row.text.replace(new RegExp('(' + opts.value + ')', 'i'), '<b>$1</b>'));
      if (row.style == 'divider') {
        row_text.css('border-top', '1px #CCC solid');
      }
      row_link.append(row_text);
    }
    
    fuzzy.html(table);
    
    autocode.initState();
    
    autocode.resize.fuzzy();
  }
};;autocode.ga = {
  send: function(type, href) {
    if (typeof ga !== 'undefined') {
      ga('send', type, href);
    }
  }
};;autocode.hint = {
  locked: false,
  timer: null,
  close: function(opts) {
    opts = opts || {};
    if (opts.animated === false) {
      $('#hint, #hint-arrow').remove();
    } else {
      $('#hint, #hint-arrow').remove();
      /*
      $('#hint, #hint-arrow').fadeOut(function() {
        $('#hint, #hint-arrow').remove();
      });
      */
    }
  },
  init: function() {
    $('*[data-hint]').each(function() {
      $(this).bind({
        mouseenter: function() {
          if (autocode.hint.locked) {
            return;
          }
          
          autocode.hint.open({
            minTop: $(this).data('hint-min-top'),
            target: $(this),
            text: $(this).data('hint')
          });
          
          autocode.resize.hint();
        },
        mouseleave: function() {
          if (autocode.hint.locked) {
            return;
          }
          
          autocode.hint.close();
        }
      });
    });
  },
  open: function(opts) {
    clearTimeout(autocode.hint.timer);
    
    var hint = $('#hint');
    if (!hint.length) {
      hint = $(document.createElement('div'));
      hint.attr('id', 'hint');
      $('body').append(hint);
    }
    if (opts.minTop !== undefined) {
      hint.data('minTop', opts.minTop);
    } else {
      hint.removeData('minTop');
    }
    if (opts.offsetTop !== undefined) {
      hint.data('offsetTop', opts.offsetTop);
    } else {
      hint.removeData('offsetTop');
    }
    hint.data('originalText', opts.text);
    if (opts.scrollUp) {
      hint.data('scrollUp', opts.scrollUp);
    } else {
      hint.removeData('scrollUp');
    }
    if (opts.top !== undefined) {
      hint.data('top', opts.top);
    } else {
      hint.removeData('top');
    }
    
    var hint_arrow = $('#hint-arrow');
    if (!hint_arrow.length) {
      hint_arrow = $(document.createElement('div'));
      hint_arrow.attr('id', 'hint-arrow');
      $('body').append(hint_arrow);
    }
    
    hint.data('target', opts.target);
    hint.html(opts.text);
    
    if (opts.timer) {
      autocode.hint.timer = setTimeout(function() {
        autocode.hint.close();
      }, opts.timer);
    }
  }
};;autocode.init = function() {
  autocode.api.url = autocode.url.api();
  autocode.resize.loader();
  $('#loader').animate({ opacity: 1 });
  
  autocode.console.init();
  autocode.editor.init();
  autocode.hint.init();
  autocode.shortcut.init()
  autocode.initState();
  
  $(window).bind({
    keyup: function(e) {
      if (e.keyCode == 9 || e.keyCode == 27) {
        autocode.fuzzy.close();
      }
    },
    mousedown: function(e) {
      var target = $(e.target);
      
      if (
        target.attr('id') != 'popover' && !target.parents('#popover').length
        && target.attr('id') != 'menu' && !target.parents('#menu').length
        && target.attr('id') != 'more-icon' && !target.parents('#more-icon').length
        && target.attr('id') != 'user' && !target.parents('#user').length
      ) {
        autocode.popover.close();
      }
      
      if (target.attr('id') != 'fuzzy' && !target.parents('#fuzzy').length) {
        autocode.fuzzy.close();
      }
    },
    popstate: function() {
      var hash = location.hash.split('#')[1];
      if (autocode.state[hash]) {
        autocode.state[hash]();
      }
    },
    scroll: function() {
      $(this).data('scrolled', true);
    }
  });
  
  $('#content .content-center').bind('scroll', function() {
    if ($('#fuzzy').length) {
      autocode.resize.fuzzy();
    }
    if ($('#hint').length) {
      autocode.resize.hint();
    }
  });
  
  var code = autocode.query.get('code');
  if (code) {
    history.pushState(null, null, '/');
    autocode.api.login.post({
      data: {
        code: code,
        provider: 1
      },
      error: function(data) {
        alert('Unable to login.');
        autocode.load();
      },
      success: autocode.load
    });
  } else {
    autocode.load();
  }
  
  autocode.resize.all();
};

$(window).load(autocode.init);;autocode.listener = {
  listeners: {},
  add: function(name, state, callback) {
    if (!autocode.listener.listeners[state]) {
      autocode.listener.listeners[state] = {};
    }
    autocode.listener.listeners[state][name] = callback;
  },
  remove: function(name, state) {
    if (state) {
      autocode.listener.listeners[state].splice(
        autocode.listener.listeners[state].indexOf(name),
        1
      );
      if (!Object.keys(autocode.listener.listeners[state]).length) {
        delete(autocode.listener.listeners[state]);
      }
    } else {
      var listeners;
      for (var listener_state in autocode.listener.listeners) {
        delete(autocode.listener.listeners[listener_state][name]);
        if (!Object.keys(autocode.listener.listeners[listener_state]).length) {
          delete(autocode.listener.listeners[listener_state]);
        }
      }
    }
  }
};;autocode.load = function() {
  autocode.api.auth.get({
    error: function(data) {
      console.log(data);
      alert('Unable to load Autocode. Please contact support.');
    },
    success: function(data) {
      autocode.data.auth = data;
      
      autocode.api.config.get({
        data: {
          repo: 'crystal/autocode'
        },
        error: function(data) {
          console.log(data);
          alert('Unable to load Autocode. Please contact support.');
        },
        success: function(data) {
          autocode.config = jsyaml.safeLoad(data.config);
          
          autocode.api.user.get({
            complete: function() {
              autocode.action.updateRecent();
              
              var repo = autocode.storage.get('repo');
              if (!repo) {
                repo = location.pathname.substr(1);
              }
              
              if (!repo) {
                $('#loader').fadeOut(function() {
                  $('#loader').remove();
                  $('#container').show();
                  $('#menu-project, #welcome').hide();
                  autocode.resize.all();
                  $('#container').animate({ opacity: 1 },{
                    complete: function() {
                      var config = autocode.query.get('config');
                      if (config) {
                        config = jsyaml.safeLoad(atob(config));
                        autocode.action.loadProject({
                          name: '(Untitled)',
                          config: config
                        });
                      } else {
                        $('#welcome').css({ opacity: 0 }).show().animate({ opacity: 1 });
                      }
                      
                      autocode.resize.all();
                    }
                  });
                });
                
                return;
              }
              
              $('#loader').fadeOut(function() {
                $('#loader').remove();
                $('#container').show();
                $('#welcome').hide();
                autocode.resize.all();
                $('#container').animate({
                  opacity: 1
                },{
                  complete: function() {
                    var name = repo.split('/').splice(0, 2).join('/');
                    var state = repo.split('/').splice(2).join('/');
                    
                    if (autocode.state[name]) {
                      autocode.state['tour']();
                      return;
                    }
                    
                    autocode.action.loadProject({
                      name: name,
                      callback: state ? autocode.state[state] : null
                    });
                  }
                });
              });
            },
            error: function() {
              $('#new-option, #load-option').hide();
              $('#recent').hide();
            },
            success: function(data) {
              autocode.data.user = data;
              autocode.data.user.isLoggedIn = true;
              
              $('#login-option').hide();
              
              $('#user .icon').css('background-image', 'url(' + data.avatar + ')');
              $('#user .text').text(data.username);
              
              if (location.hostname.match(/^alpha\./)) {
                autocode.ws.init();
                $('#target-icon .text').text(autocode.ws.ip);
                
              } else {
                clearInterval(autocode.data.current.timer);
                delete(autocode.data.current.timer);
                
                var initBox = function() {
                  autocode.api.box.post({
                    error: function(data) {
                      console.log(data);
                      $('#target-icon .text').text('Target not found');
                    },
                    success: function(data) {
                      if (data.ready) {
                        if (!autocode.ws.ip || autocode.ws.ip == '127.0.0.1') {
                          autocode.status.pending();

                          autocode.ws.ip = data.ip;
                          autocode.ws.init();
                        }
                        
                        $('#target-icon .text').text(data.ip);
                      }
                    }
                  });
                };
                
                initBox();
                
                autocode.data.current.timer = setInterval(initBox, 10 * 1000);
              }
            }
          });
        }
      });
    }
  });
};;autocode.loader = {
  close: function() {
    $('#loader').remove();
  },
  open: function(opts) {
    autocode.loader.close();
    
    var loader = $(document.createElement('div'));
    loader.attr('id', 'loader');
    loader.html('<img class="icon" src="images/loader.svg" />');
    $('body').append(loader);
    
    autocode.resize.all(['loader']);
    
    loader.animate({ opacity: 1 });
  }
};;autocode.mobile = {
  minWidth: 900
};;autocode.object = {
  clone: function(o) {
    return JSON.parse(JSON.stringify(o));
  },
  first: function(o) {
    for (var i in o) {
      return o[i];
    }
  },
  firstKey: function(o) {
    for (var i in o) {
      return i;
    }
  },
  last: function(o) {
    var v;
    for (var i in o) {
      v = o[i];
    }
    return v;
  },
  lastKey: function(o) {
    var k;
    for (var i in o) {
      k = i;
    }
    return k;
  },
  sort: function(o) {
    var sorted = {}, key, a = [];

    for (key in o) {
      if (o.hasOwnProperty(key)) {
        a.push(key);
      }
    }
    a.sort();

    for (key = 0; key < a.length; key++) {
      sorted[a[key]] = o[a[key]];
    }
    
    return sorted;
  }
};;autocode.popover = {
  close: function() {
    $('#popover').remove();
  },
  open: function(opts) {
    var popover = $(document.createElement('div'));
    popover.attr('id', 'popover');
    if (opts.bottom !== undefined) {
      popover.css('bottom', opts.bottom);
    }
    if (opts.left !== undefined) {
      popover.css('left', opts.left);
    }
    if (opts.right !== undefined) {
      popover.css('right', opts.right);
    }
    if (opts.top !== undefined) {
      popover.css('top', opts.top);
    }
    popover.data('target', opts.target);
    if (!opts.content) {
      switch (opts.style) {
        case 'table':
          opts.content = '<div class="table">';
          var row_action;
          for (var row_i = 0; row_i < opts.rows.length; row_i++) {
            if (opts.rows[row_i].state) {
              row_action = ' href="' + opts.rows[row_i].state + '"';
            } else if (typeof(opts.rows[row_i].action) == 'object') {
              row_action = ' onclick="autocode.action.' + opts.rows[row_i].action.name + '(' + JSON.stringify(opts.rows[row_i].action.data) + ')"';
            } else if (typeof(opts.rows[row_i].action) == 'string') {
              row_action = ' onclick="autocode.action.' + opts.rows[row_i].action + '()"';
            } else {
              row_action = '';
            }
            
            opts.content += '<a' + row_action + '>'
                + '<span class="icon ' + opts.rows[row_i].icon + '"' + (opts.rows[row_i].style == 'divider' ? ' style="border-top: 1px #CCC solid"' : '') + '></span>'
                + '<span class="text"' + (opts.rows[row_i].style == 'divider' ? ' style="border-top: 1px #CCC solid"' : '') + '>' + opts.rows[row_i].text + '</span>'
              + '</a>';
          }
          opts.content += '</div>';
          break;
      }
    }
    popover.html(opts.content);
    popover.hide();
    $('body').append(popover);
    popover.slideDown();
  },
  toggle: function(opts) {
    if ($('#popover').length) {
      var target = $('#popover').data('target')
      $('#popover').remove();
      if (opts.target.attr('id') != target.attr('id')) {
        autocode.popover.open(opts);
      }
    } else {
      autocode.popover.open(opts);
    }
  }
};;autocode.popup = {
  close: function() {
    $('#popup, #overlay').remove();
  },
  error: function(msg) {
    if (msg === false) {
      $('#popup .error').hide();
    } else {
      $('#popup .error').text(msg).show();
    }
    
    autocode.resize.all();
    
    return false;
  },
  open: function(opts) {
    autocode.popup.close();
    
    var popup = $(document.createElement('div'));
    popup.attr('id', 'popup');
    popup.hide();
    
    var html = '';
    if (opts.title) {
      html += '<div class="title"' + (opts.content ? '' : ' style="padding-bottom: 0px"') + '>' + opts.title + '</div>';
    }
    html += '<div class="error"></div>';
    if (!opts.content) {
      switch (opts.style) {
        case 'table':
          opts.content = '<div class="table">';
          var row_action;
          for (var row_i = 0; row_i < opts.rows.length; row_i++) {
            if (opts.rows[row_i].state) {
              row_action = ' href="' + opts.rows[row_i].state + '"';
            } else if (typeof(opts.rows[row_i].action) == 'object') {
              row_action = ' onclick="autocode.action.' + opts.rows[row_i].action.name + '(' + JSON.stringify(opts.rows[row_i].action.data) + ')"';
            } else if (typeof(opts.rows[row_i].action) == 'string') {
              row_action = ' onclick="autocode.action.' + opts.rows[row_i].action + '()"';
            } else {
              row_action = '';
            }
            
            opts.content += '<a' + row_action + '>'
                + '<span class="icon' + (!opts.rows[row_i].icon.match(/^http/) ? ' ' + opts.rows[row_i].icon : '') + '" style="' + (!!opts.rows[row_i].icon.match(/^http/) ? 'background-image: url(' + opts.rows[row_i].icon + ');' : '') + (opts.rows[row_i].style == 'divider' ? 'border-top: 1px #CCC solid' : '') + '"></span>'
                + '<span class="text"' + (opts.rows[row_i].style == 'divider' ? ' style="border-top: 1px #CCC solid"' : '') + '>' + opts.rows[row_i].text + '</span>'
              + '</a>';
          }
          opts.content += '</div>';
          break;
      }
    }
    if (opts.content) {
      html += '<div class="content">' + opts.content + '</div>';
    }
    popup.html(html);
    
    $('body').append(popup);
    
    if (opts.overlay !== false) {
      var overlay = $(document.createElement('div'));
      overlay.attr('id', 'overlay');
      overlay.click(function() {
        autocode.popup.close();
      });
      overlay.hide();
      $('body').append(overlay);
      $('#overlay').fadeIn();
    }
    
    $('#popup').css('opacity', 0.01).show();
    $('#popup').animate({ opacity: 1 }, {
      complete: function() {
        $('#popup input, #popup textarea').first().focus().keyup();
        
        if (opts.complete) {
          opts.complete();
        }
      }
    });
    
    autocode.resize.all(['fuzzy', 'overlay', 'popup']);
    
    autocode.initState();
  }
};;autocode.query = {
  build: function(params) {
    return $.param(params);
  },
  get: function(variable, search) {
    var query = search ? search : window.location.search.substring(1);
    var vars = query.split('&');
    var data = {};
    for (var i = 0; i < vars.length; i++) {
      var pair = vars[i].split('=');
      if (variable && decodeURIComponent(pair[0]) == variable) {
        return decodeURIComponent(pair[1]);
      }
    }
  },
  search: function(search) {
    var query = search ? search : window.location.search.substring(1);
    var vars = query.split('&');
    var data = {};
    for (var i = 0; i < vars.length; i++) {
      var pair = vars[i].split('=');
      data[decodeURIComponent(pair[0])] = decodeURIComponent(pair[1]);
    }
    return data;
  }
};;autocode.resize = {
  all: function(resize) {
    if (typeof(resize) == 'string') {
      resize = [resize];
    }
    
    var parts = [
      'body',
      'content',
      'config',
      'exports',
      'exportsInit',
      'footer',
      'header',
      'hint',
      'init',
      'interfacesInit',
      'loader',
      'output',
      'outputs',
      'outputInit',
      'outputsInit',
      'overlay',
      'popup',
      'fuzzy',
      'welcome'
    ];
    
    var part;
    for (var i = 0; i < parts.length; i++) {
      part = parts[i];
      if (resize && resize.indexOf(part) === -1) {
        continue;
      }
      autocode.resize[part]();
    }
  }
};

$(window).resize(function() { autocode.resize.all(); });;autocode.shortcut = {
  init: function() {
    $(window).bind('keydown', function(e) {
      if (e.keyCode == 27) {
        if ($('#popup').length) {
          autocode.action.closePopup();
        }
        return;
      } else if (!e.ctrlKey || !e.shiftKey) {
        return;
      }
      
      switch (e.keyCode) {
        // 1
        case 49: {
          autocode.state['overview']();
          break;
        }
        // 2
        case 50: {
          autocode.state['imports']();
          break;
        }
        // 3
        case 51: {
          autocode.state['config']();
          break;
        }
        // 4
        case 52: {
          autocode.state['output']();
          break;
        }
        // b
        case 66: {
          autocode.action.build();
          break;
        }
        // c
        case 67: {
          autocode.action.toggleConsole();
          break;
        }
        // g
        case 71: {
          autocode.action.generate();
          break;
        }
        // h
        case 72: {
          autocode.action.halt();
          break;
        }
        // l
        case 76: {
          autocode.action.clean();
          break;
        }
        // m
        case 77: {
          autocode.action.commitProject();
          break;
        }
        // n
        case 78: {
          autocode.action.newProject();
          break;
        }
        // o
        case 79: {
          autocode.action.loadProject();
          break;
        }
        // r
        case 82: {
          autocode.action.run();
          break;
        }
        // s
        case 83: {
          autocode.action.saveProject();
          break;
        }
        // t
        case 84: {
          autocode.action.reset();
          break;
        }
        // u
        case 85: {
          autocode.action.update();
          break;
        }
        // w
        case 87: {
          autocode.action.closeProject();
          break;
        }
      }
      
      autocode.resize.all();
      
      return false;
    });
  }
};;autocode.status = {
  seconds: 0,
  timer: null,
  offline: function() {
    $('#status-icon').removeClass('online pending processing');
    $('#status-icon').data('hint', 'Offline');
    $('#target-icon .target-icon').removeClass('loading');
    
    autocode.status.startTimer();
  },
  online: function() {
    $('#status-icon').removeClass('pending processing').addClass('online');
    $('#status-icon').data('hint', 'Online');
    $('#target-icon .target-icon').removeClass('loading');
    
    autocode.status.startTimer();
  },
  pending: function() {
    $('#status-icon').removeClass('online processing').addClass('pending');
    $('#status-icon').data('hint', 'Pending');
    $('#target-icon .target-icon').addClass('loading');
    
    autocode.status.startTimer();
  },
  processing: function() {
    $('#status-icon').removeClass('online pending').addClass('processing');
    $('#status-icon').data('hint', 'Processing');
    $('#target-icon .target-icon').addClass('loading');
    
    autocode.status.startTimer();
  },
  clearTimer: function() {
    if (autocode.status.timer) {
      clearInterval(autocode.status.timer);
      delete(autocode.status.timer);
    }
    
    autocode.status.seconds = 0;
  },
  startTimer: function() {
    autocode.status.clearTimer();
    autocode.status.timer = setInterval(autocode.status.updateTimer, 1000);
  },
  updateTimer: function() {
    autocode.status.seconds++;
    if ($('#hint').is(':visible') && $('#hint').data('target') && $('#hint').data('target').attr('id') == 'status-icon') {
      $('#hint').text($('#status-icon').data('hint') + ' (' + autocode.status.seconds + ')');
    }
  }
};;autocode.storage = {
  get: function(name, default_item) {
    var item = localStorage.getItem(name);
    if (item && item.match(/^json:\/\//)) {
      item = JSON.parse(item.substr(7));
    }
    return item || default_item;
  },
  remove: function(name, item) {
    return localStorage.removeItem(name);
  },
  set: function(name, item) {
    if (typeof(item) == 'object') {
      item = 'json://' + JSON.stringify(item);
    }
    return localStorage.setItem(name, item);
  }
};;autocode.string = {
  capitalize: function(str) {
    return str.substr(0, 1).toUpperCase() + str.substr(1);
  },
  escape: function(str) {
    return str.replace(/"/g, '&quot;');
  }
};;autocode.unload = {
  callback: function() {
    return 'You will lose any unsaved changes';
  },
  disable: function() {
    if (location.hostname.match(/^alpha/)) {
      return;
    }
    $(window).unbind('beforeunload', autocode.unload.callback);
  },
  enable: function() {
    if (location.hostname.match(/^alpha/)) {
      return;
    }
    $(window).bind('beforeunload', autocode.unload.callback);
  }
};;autocode.url = {
  api: function(uri) {
    if (location.host.match(/^alpha/)) {
      return 'http://alpha.api.autocode.run:3000/' + (uri ? uri : '');
    } else {
      return 'https://api.autocode.run/' + (uri ? uri : '');
    }
  }
};;autocode.ws = {
  ip: '127.0.0.1',
  io: null,
  init: function() {
    if (!autocode.ws.ip) {
      return;
    }
    
    $('#console .content').append('<div>Connecting...</div>');
    if (autocode.data.current.pin) {
      $('#console .content').scrollTop($('#console .content')[0].scrollHeight);
    }
    
    autocode.ws.io = io('http://' + autocode.ws.ip + ':12345');
    autocode.ws.io.on('connect', function(socket) {
      $('#console .content').append('<div>Connected.</div>');
      if (autocode.data.current.pin) {
        $('#console .content').scrollTop($('#console .content')[0].scrollHeight);
      }
      autocode.status.online();
      
      $('#usage-on').show();
      $('#usage-off').hide();
    });
    autocode.ws.io.on('contents', function(data) {
      autocode.action.loadFile({
        file: data.file,
        value: data.contents
      });
    });
    autocode.ws.io.on('error', function(data) {
      alert(data.error);
    });
    autocode.ws.io.on('files', function(data) {
      autocode.data.output = data;
      autocode.state['output']();
    });
    autocode.ws.io.on('message', function(data) {
      if (data.processing === true) {
        autocode.status.processing();
        $('#build-icon, #generate-icon, #run-icon, #update-icon').addClass('disabled');
      } else if (data.processing === false) {
        autocode.status.online();
        $('#build-icon, #generate-icon, #run-icon, #update-icon').removeClass('disabled');
      }
      
      if (data.running === true) {
        autocode.data.current.running = true;
        $('#run-icon').addClass('stop');
      } else if (data.running === false) {
        autocode.data.current.running = false;
        $('#run-icon').removeClass('stop');
      }
      
      if (!data.data) {
        return;
      }
      data = data.data.split("\n");
      for (var i = 0; i < data.length; i++) {
        $('#console .content').append('<div>' + (data[i].length ? data[i] : '&nbsp;') + '</div>');
        if (autocode.data.current.pin) {
          $('#console .content').scrollTop($('#console .content')[0].scrollHeight);
        }
      }
    });
    autocode.ws.io.on('disconnect', function(data) {
      $('#console .content').append('<div>Disconnected.</div>');
      if (autocode.data.current.pin) {
        $('#console .content').scrollTop($('#console .content')[0].scrollHeight);
      }
      autocode.status.pending();
      
      $('#usage-on').hide();
      $('#usage-off').show();
    });
  }
};;;;autocode.action.addCommand = function() {
  autocode.popup.open({
    title: 'Loading...'
  });
  autocode.popover.close();
  
  new formulator({
    formula: 'formulas/forms/AddCommand.json',
    xhr: true,
    ready: function(form) {
      form.fields.command.autocomplete = false;
      
      autocode.popup.open({
        title: 'Add Command',
        content: form.toString()
      });
    },
    submit: function(data) {
      var data = {};
      $('#popup input, #popup select, #popup textarea').each(function() {
        data[$(this).attr('name')] = $(this).val();
      });
      
      if (!autocode.project.scripts) {
        autocode.project.scripts = {};
      }
      if (!autocode.project.scripts[autocode.data.current.script]) {
        autocode.project.scripts[autocode.data.current.script] = [];
      }
      
      var script = {
        description: data.description,
        command: data.command
      };
      if (data.path.length) {
        script.path = data.path;
      }
      
      autocode.project.scripts[autocode.data.current.script].push(script);
      
      $('#popup, #overlay').fadeOut(function() {
        autocode.popup.close();
      });
      
      autocode.action.loadScript();
      
      return false;
    }
  });
};;autocode.action.addCommandTitle = function() {
  autocode.popup.open({
    title: 'Loading...'
  });
  autocode.popover.close();
  
  new formulator({
    formula: 'formulas/forms/AddCommandTitle.json',
    xhr: true,
    ready: function(form) {
      autocode.popup.open({
        title: 'Add Title',
        content: form.toString()
      });
    },
    submit: function(data) {
      var data = {};
      $('#popup input, #popup select, #popup textarea').each(function() {
        data[$(this).attr('name')] = $(this).val();
      });
      
      if (!autocode.project.scripts) {
        autocode.project.scripts = {};
      }
      if (!autocode.project.scripts[autocode.data.current.script]) {
        autocode.project.scripts[autocode.data.current.script] = [];
      }
      
      var script = {
        title: data.title
      };
      
      autocode.project.scripts[autocode.data.current.script].push(script);
      
      $('#popup, #overlay').fadeOut(function() {
        autocode.popup.close();
      });
      
      autocode.action.loadScript();
      
      return false;
    }
  });
};;autocode.action.addExport = function() {
  autocode.popup.open({
    title: 'Loading...'
  });
  autocode.popover.close();
  
  new formulator({
    formula: 'formulas/forms/AddExport.json',
    xhr: true,
    ready: function(form) {
      form.fields.type.autocomplete = false;
      form.fields.type.focus = function(o) {
        $(o).trigger('keyup');
      };
      form.fields.type.keyup = function() {
        var rows = [], type_name, types = autocode.data.exportTypes;
        for (var i in types) {
          type_name = types[i];
          if (type_name.match(new RegExp(value, 'i'))) {
            rows.push({
              action: {
                name: 'addExportName',
                data: { name: autocode.string.capitalize(type_name) }
              },
              icon: type_name.toLowerCase() + '-icon black',
              text: autocode.string.capitalize(type_name)
            });
          }
        }
        
        autocode.fuzzy.open({
          rows: rows,
          target: $('#popup input[name="type"]'),
          value: value
        });
      };
      
      form.fields.engine.autocomplete = false;
      form.fields.engine.focus = function(o) {
        $(o).trigger('keyup');
      };
      form.fields.engine.keyup = function() {
        var engine, engines = [];
        for (var engine_name in autocode.data.engines) {
          engine = autocode.data.engines[engine_name];
          if (engine_name.match(new RegExp(value, 'i'))) {
            engines.push({
              action: {
                name: 'addExportEngine',
                data: { name: engine_name }
              },
              icon: 'https://cdn.rawgit.com/crystal/' + engine_name.split('.')[0] + '/master/.autocode/icon.svg',
              text: engine_name
            });
          }
        }
        
        autocode.fuzzy.open({
          rows: engines,
          target: $('#popup input[name="engine"]'),
          value: value
        });
      };
      
      form.fields.format.autocomplete = false;
      form.fields.format.focus = function(o) {
        $(o).trigger('keyup');
      };
      form.fields.format.keyup = function() {
        var rows = [], type_name, types = autocode.data.formats;
        for (var i in types) {
          type_name = types[i];
          if (type_name.match(new RegExp(value, 'i'))) {
            rows.push({
              action: {
                name: 'addExportName',
                data: { name: type_name }
              },
              icon: type_name.toLowerCase() + '-icon black',
              text: type_name
            });
          }
        }
        
        autocode.fuzzy.open({
          rows: rows,
          target: $('#popup input[name="format"]'),
          value: value
        });
      };
      
      autocode.popup.open({
        title: 'Add Export',
        content: form.toString()
      });
      
      $('#popup input[name="engine"]').parents('.field').hide();
      $('#popup input[name="filename"]').parents('.field').hide();
      $('#popup input[name="format"]').parents('.field').hide();
      $('#popup input[name="schema"]').parents('.field').hide();
      $('#popup textarea[name="template"]').parents('.field').hide();
      
      autocode.resize.popup();
    },
    submit: function(data) {
      var data = {};
      $('#popup input, #popup select, #popup textarea').each(function() {
        var value = $(this).val();
        if (value.length) {
          data[$(this).attr('name')] = $(this).val();
        }
      });
      
      if (data.type) {
        data.type = data.type.toLowerCase();
      }
      
      var name = data.name;
      delete(data.name);
      if (!autocode.project.exports) {
        autocode.project.exports = {};
      }
      autocode.project.exports[name] = data;
      
      // sort exports
      autocode.project.exports = autocode.object.sort(autocode.project.exports);
      
      $('#popup, #overlay').fadeOut(function() {
        autocode.popup.close();
      });
      
      autocode.state['exports']();
      
      return false;
    }
  });
};;autocode.action.addExportEngine = function(opts) {
  $('#popup input[name="engine"]').val(opts.name);
  autocode.fuzzy.close();
};;autocode.action.addExportName = function(opts) {
  $('#popup input[name="type"]').val(opts.name);
  
  switch (true) {
    case ['engine','helper','processor','schema','spec','transformer'].indexOf(opts.name.toLowerCase()) !== -1: {
      $('#popup input[name="engine"]').parents('.field').hide();
      $('#popup input[name="filename"]').parents('.field').hide();
      $('#popup input[name="format"]').parents('.field').hide();
      $('#popup input[name="schema"]').parents('.field').hide();
      $('#popup textarea[name="template"]').parents('.field').hide();
      break;
    }
    case opts.name.toLowerCase() == 'generator': {
      $('#popup input[name="engine"]').parents('.field').show();
      $('#popup input[name="filename"]').parents('.field').show();
      $('#popup input[name="format"]').parents('.field').show();
      $('#popup input[name="schema"]').parents('.field').show();
      $('#popup textarea[name="template"]').parents('.field').show();
      break;
    }
  }
  
  autocode.fuzzy.close();
  
  autocode.resize.popup();
};;autocode.action.addExportSchema = function(opts) {
  $('#popup input[name="schema"]').val(opts.name);
  autocode.fuzzy.close();
};;autocode.action.addImport = function(opts) {
  autocode.fuzzy.close();
  
  if (!autocode.project.imports) {
    autocode.project.imports = {};
  }
  
  autocode.project.imports[opts.repo] = 'Loading...';
  
  autocode.state['imports']({ disableSelected: true });
  $('#imports-search').val('');
  $('#imports-content-readme').text('');
  
  var add = function() {
    autocode.api.releases.get({
      data: {
        repo: opts.repo
      },
      success: function(data) {
        autocode.project.imports[opts.repo] = '~' + data[0].substr(1);
        
        autocode.api.config.get({
          data: {
            repo: opts.repo
          },
          success: function(data) {
            var imported = this.url.split('?')[1];
            imported = autocode.query.search(imported);
            
            var import_name = imported.repo.split('/');
            if (import_name.length > 1) {
              import_name = import_name[1];
            } else {
              import_name = import_name[0];
            }
            
            data.config = jsyaml.safeLoad(data.config)
            
            autocode.imports[import_name] = data.config;
            
            // sort imports
            autocode.project.imports = autocode.object.sort(autocode.project.imports);
            
            if (data.config.exports) {
              for (var export_name in data.config.exports) {
                switch (data.config.exports[export_name].type) {
                  case 'generator': {
                    autocode.data.generators[imported.repo.split('/')[1] + '.' + export_name] = JSON.parse(JSON.stringify(data.config.exports[export_name]));
                    break;
                  }
                }
              }
            }
          }
        });
        
        autocode.state['imports']({ disableSelected: true });
        if ($(window).width() > autocode.mobile.minWidth) {
          autocode.action.loadImport({ repo: opts.repo });
        }
        
        $('#imports-init').fadeOut(function() {
          $('#imports-content-container').fadeIn();
        });
      }
    });
  };
  
  if ($('#popup').length) {
    $('#popup, #overlay').fadeOut(add);
  } else {
    add();
  }
};;autocode.action.addInterface = function() {
  autocode.popup.open({
    title: 'Loading...'
  });
  autocode.popover.close();
  
  new formulator({
    formula: 'formulas/forms/AddInterface.json',
    xhr: true,
    ready: function(form) {
      autocode.popup.open({
        title: 'Add Interface',
        content: form.toString()
      });
    },
    submit: function(data) {
      var data = {};
      $('#popup input, #popup select, #popup textarea').each(function() {
        data[$(this).attr('name')] = $(this).val();
      });
      
      var interface_data = {
        name: data.name,
        port: data.port
      };
      
      if (!autocode.project.interfaces) {
        autocode.project.interfaces = [];
      }
      autocode.project.interfaces.push(interface_data);
      
      $('#popup, #overlay').fadeOut(function() {
        autocode.popup.close();
      });
      
      autocode.state['interfaces']();
      
      return false;
    }
  });
};;autocode.action.addOutput = function() {
  autocode.popup.open({
    title: 'Loading...'
  });
  autocode.popover.close();
  
  new formulator({
    formula: 'formulas/forms/AddOutput.json',
    xhr: true,
    ready: function(form) {
      form.fields.filename.autocomplete = false;
      form.fields.generator.autocomplete = false;
      
      form.fields.generator.keyup = function() {
        var value = $('#popup input[name="generator"]').val();
        
        var current_generators = autocode.current.generators();
        var generator, generator_icon, generators = [];
        for (var generator_name in current_generators) {
          generator = current_generators[generator_name];
          if (generator_name.split('.').length > 1) {
            generator_icon = 'https://cdn.rawgit.com/crystal/' + generator_name.split('.')[0] + '/master/.autocode/icon.svg';
          } else {
            generator_icon = 'project-icon black';
          }
          if (generator_name.match(new RegExp(value, 'i'))) {
            generators.push({
              action: {
                name: 'addOutputGenerator',
                data: {
                  name: generator_name
                }
              },
              icon: generator_icon,
              text: generator_name
            });
          }
        }
        autocode.fuzzy.open({
          rows: generators,
          target: $('#popup input[name="generator"]'),
          value: value
        });
        
        $('#popup input[name="filename"]').attr('placeholder', current_generators[value] && current_generators[value].filename ? current_generators[value].filename : '');
      };
      
      autocode.popup.open({
        title: 'Add Output',
        content: form.toString()
      });
    },
    submit: function(data) {
      var data = {};
      $('#popup input, #popup select, #popup textarea').each(function() {
        data[$(this).attr('name')] = $(this).val();
      });
      
      // validate generator
      var current_generators = autocode.current.generators();
      if (!current_generators[data.generator]) {
        autocode.popup.error('Generator does not exist.');
        return false;
      }
      
      // normalize filename
      if (!data.filename) {
        data.filename = current_generators[data.generator].filename;
      }
      
      // validate filename
      var output;
      for (var output_i in autocode.project.outputs) {
        output = autocode.project.outputs[output_i];
        if (
          (output.filename && output.filename == data.filename)
          ||
          (!output.filename && current_generators[output.generator] && current_generators[output.generator].filename == data.filename)
        ) {
          autocode.popup.error('Filename must be unique.');
          return false;
        }
      }
      
      if (data.filename == current_generators[data.generator].filename) {
        delete(data.filename);
      }
      
      if (!autocode.project.outputs) {
        autocode.project.outputs = [];
      }
      autocode.project.outputs.push(data);
      
      $('#popup, #overlay').fadeOut(function() {
        autocode.popup.close();
      });
      
      autocode.state['outputs']();
      autocode.action.loadOutput({ output: autocode.project.outputs.length-1 });
      
      autocode.resize.all();
      
      return false;
    }
  });
};;autocode.action.addOutputGenerator = function(opts) {
  $('#popup input[name="generator"]').val(opts.name);
  $('#popup input[name="filename"]').attr('placeholder', autocode.data.generators[opts.name] && autocode.data.generators[opts.name].filename ? autocode.data.generators[opts.name].filename : '');
  autocode.fuzzy.close();
};;autocode.action.addProperty = function() {
  new formulator({
    formula: 'formulas/forms/AddProperty.json',
    xhr: true,
    ready: function(form) {
      form.fields.property.autocomplete = false;
      
      form.fields.property.keyup = function() {
        var value = $('#popup input[name="property"]').val();
        
        var property, properties = [];
        for (var property_name in autocode.object.sort(autocode.data.generators[autocode.data.current.generator].schema.properties)) {
          property = autocode.data.generators[autocode.data.current.generator].schema.properties[property_name];
          if (property_name.match(new RegExp(value, 'i'))) {
            properties.push({
              state: 'outputs/property/add/choose?name=' + property_name,
              text: property_name
            });
          }
        }
        autocode.fuzzy.open({
          rows: properties,
          target: $('#popup input[name="property"]'),
          value: value
        });
      };
      
      autocode.popup.open({
        title: 'Add Property',
        content: form.toString()
      });
    },
    submit: function() {
      var data = {};
      $('#popup input, #popup select, #popup textarea').each(function() {
        data[$(this).attr('name')] = $(this).val();
      });
      
      // validate name
      if (!autocode.data.generators[autocode.data.current.generator].schema.properties[data.property]) {
        autocode.popup.error('Property does not exist.');
        return false;
      }
      
      // validate value
      if (autocode.data.generators[autocode.data.current.generator].schema.properties[data.property].type == 'boolean' && data.value != 'true' && data.value != 'false') {
        autocode.popup.error('Value must be a boolean.');
        return false;
      } else if (autocode.data.generators[autocode.data.current.generator].schema.properties[data.property].type == 'number' && !data.value.match(/^\d+$/)) {
        autocode.popup.error('Value must be a number.');
        return false;
      }
      
      if (autocode.data.generators[autocode.data.current.generator].schema.properties[data.property].type == 'boolean') {
        if (data.value == 'true') {
          data.value = true;
        } else {
          data.value = false;
        }
      }
      
      if (!autocode.project.outputs[autocode.data.current.output].spec) {
        autocode.project.outputs[autocode.data.current.output].spec = {};
      }
      autocode.project.outputs[autocode.data.current.output].spec[data.property] = data.value;
      
      autocode.popup.close();
      
      autocode.action.loadOutput({ output: autocode.data.current.output });
      
      return false;
    }
  });
};;autocode.action.addScript = function() {
  autocode.popup.open({
    title: 'Loading...'
  });
  autocode.popover.close();
  
  new formulator({
    formula: 'formulas/forms/AddScript.json',
    xhr: true,
    ready: function(form) {
      form.fields.script.autocomplete = false;
      
      form.fields.script.keyup = function() {
        var rows = [], script_name, scripts = ['build','run','stop'];
        for (var i in scripts) {
          script_name = scripts[i];
          if (script_name.match(new RegExp(value, 'i'))) {
            rows.push({
              action: {
                name: 'addScriptName',
                data: { name: script_name }
              },
              icon: script_name + '-icon black',
              text: script_name
            });
          }
        }
        
        autocode.fuzzy.open({
          rows: rows,
          target: $('#popup input[name="script"]'),
          value: value
        });
      };
      
      autocode.popup.open({
        title: 'Add Script',
        content: form.toString()
      });
    },
    submit: function(data) {
      var data = {};
      $('#popup input, #popup select, #popup textarea').each(function() {
        data[$(this).attr('name')] = $(this).val();
      });
      
      // validate script
      if (['build','run','stop'].indexOf(data.script) === -1) {
        autocode.popup.error('Script does not exist.');
        return false;
      }
      
      if (!autocode.project.scripts) {
        autocode.project.scripts = {};
      }
      autocode.project.scripts[data.script] = [];
      
      $('#popup, #overlay').fadeOut(function() {
        autocode.popup.close();
      });
      
      autocode.state['scripts']();
      
      return false;
    }
  });
};;autocode.action.addScriptName = function(opts) {
  $('#popup input[name="script"]').val(opts.name);
  autocode.fuzzy.close();
};;autocode.action.build = function() {
  if (autocode.data.current.tab == 'config') {
    var config = $('#config-content .CodeMirror')[0].CodeMirror.getValue();
    autocode.project = jsyaml.safeLoad(config);
  }
  
  if (!autocode.project.outputs || !autocode.project.imports) {
    if (!autocode.project.imports) {
      autocode.popup.open({
        title: 'Unable to Build Project',
        content: '<div style="padding-bottom: 10px">You haven\'t imported any modules yet.</div>'
          + '<a class="button" href="imports">Import a Module</a> <button class="secondary" onclick="autocode.action.closePopup()">Cancel</button>'
      });
    } else {
      autocode.popup.open({
        title: 'Unable to Build Project',
        content: '<div style="padding-bottom: 10px">You haven\'t added any outputs to your Autocode configuration.</div>'
          + '<a class="button" href="config">Add an Output</a> <button class="secondary" onclick="autocode.action.closePopup()">Cancel</button>'
      });
    }
    
    return;
  }
  
  autocode.ws.io.emit('build', {
    config: autocode.project,
    project: autocode.repo.split('/')[1],
    user: autocode.repo.split('/')[0]
  });
};;autocode.action.clean = function(opts) {
  opts = opts || {};
  
  if (opts.confirmed) {
    // close popup
    autocode.popup.close();
    
    // clean project
    autocode.ws.io.emit('clean', {
      config: autocode.project,
      project: autocode.repo.split('/')[1],
      user: autocode.repo.split('/')[0]
    });
    
    return;
  }
  
  // close popover
  autocode.popover.close();
  
  // open popup
  autocode.popup.open({
    title: 'Are you sure?',
    content: '<div style="padding-bottom: 20px">Cleaning your project means that all of your generated files will be destroyed and need to be regenerated.</div><button class="delete" onclick="autocode.action.clean({ confirmed: true })">Yes, Do It</button> <button class="secondary" onclick="autocode.action.closePopup()">No, Nevermind</button>'
  });
};;autocode.action.clearConsole = function() {
  $('#console .content').text('');
};;autocode.action.closePopup = function() {
  autocode.popup.close();
};;autocode.action.closeProject = function(opts) {
  opts = opts || {};
  
  autocode.popover.close();
  
  var config = autocode.storage.get('config');
  if (
    opts.confirm !== true
    &&
    (
      (config && config[autocode.repo] && config[autocode.repo] != jsyaml.safeDump(autocode.project))
      ||
      ((!config || !config[autocode.repo]) && autocode.data.originalConfig != jsyaml.safeDump(autocode.project))
    )
  ) {
    autocode.popup.open({
      title: 'Close Project',
      content: '<div style="padding-bottom: 15px">Are you sure you want to close this project? <b>You will lose all unsaved changes.</b></div>'
        + '<button onclick="autocode.action.closePopup()">No, Keep It Open</button> <button class="secondary" onclick="autocode.action.closeProject({ confirm: true })">Yes, Close Project</button> <button class="secondary" onclick="autocode.action.diff()">View Unsaved Changes</button>'
    });
    return;
  }
  
  autocode.unload.disable();
  
  autocode.data.current = {};
  delete(autocode.project);
  
  autocode.popup.close();
  
  $('#menu-project').hide();
  
  $('.app, #init, header').fadeOut(function() {
    $('#welcome').fadeIn();
  });
};;autocode.action.commitProject = function(opts) {
  opts = opts || {};
  
  if (!autocode.project) {
    return;
  }

  autocode.popover.close();
  
  if (opts.confirm) {
    var data = {
      config: jsyaml.safeDump(autocode.project),
      message: $('#popup textarea[name="message"]').val(),
      repo: autocode.repo
    };
    
    autocode.popup.open({
      title: 'Committing Project and Pushing to GitHub...'
    });
    autocode.popover.close();
    
    autocode.api.config.post({
      data: data,
      error: function(data) {
        autocode.popup.open({
          title: 'Unable to Commit Project',
          content: 'Please try again or contact us at <a href="mailto:support@autocode.run">support@autocode.run</a>.'
        });
      },
      success: function(data) {
        var config = autocode.storage.get('config');
        if (!config || typeof(config) != 'object') {
          config = {};
        }
        delete(config[autocode.repo]);
        autocode.storage.set('config', config);
        autocode.data.originalConfig = jsyaml.safeDump(autocode.project);
        autocode.popup.close();
      }
    });
    return;
  }
  
  if (autocode.data.current.tab == 'config') {
    var value = $('#config-content .CodeMirror')[0].CodeMirror.getValue();
    autocode.project = jsyaml.safeLoad(value);
  }

  if (autocode.data.originalConfig == jsyaml.safeDump(autocode.project)) {
    autocode.popup.open({
      title: 'No Changes',
      content: 'There are no changes to your Autocode configuration.'
    });
    return;
  }

  var output;
  for (var output_i in autocode.project.outputs) {
    output = autocode.project.outputs[output_i];
    
    if (!autocode.data.generators[output.generator] && !autocode.project.exports[output.generator]) {
      autocode.popup.open({
        title: 'Validation Error',
        content: 'Generator does not exist for output: <b>' + output.generator + '</b>'
      });
      return;
    }
  }

  autocode.popup.open({
    title: 'Commit Project',
    content: '<div>Review your Autocode Configuration changes below before committing/pushing them to GitHub:</div><div class="diff"></div><textarea name="message" placeholder="Your commit message"></textarea><button onclick="autocode.action.commitProject({ confirm: true })">Commit Project</button>'
  });

  CodeMirror.MergeView($('#popup .diff')[0], {
    value: jsyaml.safeDump(autocode.project),
    orig: autocode.data.originalConfig,
    showDifferences: true,
    lineNumbers: true,
    mode: 'yaml',
    readOnly: true,
    revertButtons: false
  });
  
  autocode.resize.popup();
};;autocode.action.diff = function() {
  autocode.popup.open({
    title: 'Review Changes',
    content: '<div class="diff"></div><a class="button" href="project/save">Save Project</a> <button class="secondary" onclick="autocode.action.closePopup()">Close</button>'
  });
  
  CodeMirror.MergeView($('#popup .diff')[0], {
    value: jsyaml.safeDump(autocode.project),
    orig: autocode.data.originalConfig,
    lineNumbers: true,
    mode: 'yaml',
    readOnly: true,
    revertButtons: false
  });
  
  autocode.resize.popup();
};;autocode.action.editAuthorEmail = function() {
  autocode.popup.open({
    title: 'Loading...'
  });
  autocode.popover.close();
  
  new formulator({
    data: {
      email: autocode.project.author.email
    },
    formula: 'formulas/forms/Email.json',
    xhr: true,
    ready: function(form) {
      autocode.popup.open({
        title: 'Edit Author Email',
        content: form.toString()
      });
    },
    submit: function(data) {
      var data = {};
      $('#popup input, #popup select, #popup textarea').each(function() {
        data[$(this).attr('name')] = $(this).val();
      });
      
      if (data.email && data.email.length) {
        if (!autocode.project.author) {
          autocode.project.author = {};
        }
        autocode.project.author.email = data.email;
      } else {
        delete(autocode.project.author.email);
      }
      
      autocode.popup.close();
      
      autocode.state['overview/author']();
      
      return false;
    }
  });
};;autocode.action.editAuthorName = function() {
  autocode.popup.open({
    title: 'Loading...'
  });
  autocode.popover.close();
  
  new formulator({
    data: {
      name: autocode.project.author.name
    },
    formula: 'formulas/forms/Name.json',
    xhr: true,
    ready: function(form) {
      autocode.popup.open({
        title: 'Edit Author Name',
        content: form.toString()
      });
    },
    submit: function(data) {
      var data = {};
      $('#popup input, #popup select, #popup textarea').each(function() {
        data[$(this).attr('name')] = $(this).val();
      });
      
      if (data.name && data.name.length) {
        if (!autocode.project.author) {
          autocode.project.author = {};
        }
        autocode.project.author.name = data.name;
      } else {
        delete(autocode.project.author.name);
      }
      
      autocode.popup.close();
      
      autocode.state['overview/author']();
      
      return false;
    }
  });
};;autocode.action.editAuthorURL = function() {
  autocode.popup.open({
    title: 'Loading...'
  });
  autocode.popover.close();
  
  new formulator({
    data: {
      url: autocode.project.author.url
    },
    formula: 'formulas/forms/URL.json',
    xhr: true,
    ready: function(form) {
      autocode.popup.open({
        title: 'Edit Author URL',
        content: form.toString()
      });
    },
    submit: function(data) {
      var data = {};
      $('#popup input, #popup select, #popup textarea').each(function() {
        data[$(this).attr('name')] = $(this).val();
      });
      
      if (data.url && data.url.length) {
        if (!autocode.project.author) {
          autocode.project.author = {};
        }
        autocode.project.author.url = data.url;
      } else {
        delete(autocode.project.author.url);
      }
      
      autocode.popup.close();
      
      autocode.state['overview/author']();
      
      return false;
    }
  });
};;autocode.action.editCommand = function(opts) {
  opts = opts || {};
  
  autocode.data.current.command = opts.index;
  
  autocode.popup.open({
    title: 'Loading...'
  });
  autocode.popover.close();
  
  new formulator({
    formula: 'formulas/forms/EditCommand.json',
    xhr: true,
    ready: function(form) {
      form.data = autocode.data.current.commands[opts.index];
      form.fields.command.autocomplete = false;
      
      autocode.popup.open({
        title: 'Edit Command',
        content: form.toString()
      });
    },
    submit: function(data) {
      var data = {};
      $('#popup input, #popup select, #popup textarea').each(function() {
        data[$(this).attr('name')] = $(this).val();
      });
      
      if (!autocode.project.scripts) {
        autocode.project.scripts = {};
      }
      if (!autocode.project.scripts[autocode.data.current.script]) {
        autocode.project.scripts[autocode.data.current.script] = [];
      }
      
      var script = {
        description: data.description,
        command: data.command
      };
      if (data.path.length) {
        script.path = data.path;
      }
      
      autocode.project.scripts[autocode.data.current.script][autocode.data.current.command] = script;
      
      $('#popup, #overlay').fadeOut(function() {
        autocode.popup.close();
      });
      
      autocode.action.loadScript();
      
      return false;
    }
  });
};;autocode.action.editCommandTitle = function(opts) {
  opts = opts || {};
  
  autocode.data.current.command = opts.index;
  
  autocode.popup.open({
    title: 'Loading...'
  });
  autocode.popover.close();
  
  new formulator({
    formula: 'formulas/forms/EditCommandTitle.json',
    xhr: true,
    ready: function(form) {
      form.data = autocode.data.current.commands[opts.index];
      
      autocode.popup.open({
        title: 'Edit Title',
        content: form.toString()
      });
    },
    submit: function(data) {
      var data = {};
      $('#popup input, #popup select, #popup textarea').each(function() {
        data[$(this).attr('name')] = $(this).val();
      });
      
      if (!autocode.project.scripts) {
        autocode.project.scripts = {};
      }
      if (!autocode.project.scripts[autocode.data.current.script]) {
        autocode.project.scripts[autocode.data.current.script] = [];
      }
      
      var script = {
        title: data.title
      };
      
      autocode.project.scripts[autocode.data.current.script][autocode.data.current.command] = script;
      
      $('#popup, #overlay').fadeOut(function() {
        autocode.popup.close();
      });
      
      autocode.action.loadScript();
      
      return false;
    }
  });
};;autocode.action.editDescription = function() {
  autocode.popup.open({
    title: 'Loading...'
  });
  autocode.popover.close();
  
  new formulator({
    data: autocode.project,
    formula: 'formulas/forms/Description.json',
    xhr: true,
    ready: function(form) {
      autocode.popup.open({
        title: 'Edit Project Description',
        content: form.toString()
      });
    },
    submit: function(data) {
      var data = {};
      $('#popup input, #popup select, #popup textarea').each(function() {
        data[$(this).attr('name')] = $(this).val();
      });
      
      if (data.description && data.description.length) {
        autocode.project.description = data.description;
      } else {
        delete(autocode.project.description);
      }
      
      autocode.popup.close();
      
      autocode.state['overview/general']();
      
      return false;
    }
  });
};;autocode.action.editExportDescription = function() {
  autocode.popup.open({
    title: 'Loading...'
  });
  autocode.popover.close();
  
  new formulator({
    data: {
      description: autocode.project.exports[autocode.data.current.export].description
    },
    formula: 'formulas/forms/Description.json',
    xhr: true,
    ready: function(form) {
      autocode.popup.open({
        title: 'Edit Export Description',
        content: form.toString()
      });
    },
    submit: function(data) {
      var data = {};
      $('#popup input, #popup select, #popup textarea').each(function() {
        data[$(this).attr('name')] = $(this).val();
      });
      
      autocode.project.exports[autocode.data.current.export].description = data.description;
      
      autocode.action.loadExport({ repo: data.name });
      
      autocode.popup.close();
      
      return false;
    }
  });
};;autocode.action.editExportEngine = function() {
  autocode.popup.open({
    title: 'Loading...'
  });
  autocode.popover.close();
  
  new formulator({
    data: {
      engine: autocode.project.exports[autocode.data.current.export].engine
    },
    formula: 'formulas/forms/Engine.json',
    xhr: true,
    ready: function(form) {
      form.fields.engine.autocomplete = false;
      form.fields.engine.focus = function(o) {
        $(o).trigger('keyup');
      };
      form.fields.engine.keyup = function() {
        var engine, engines = [];
        for (var engine_name in autocode.data.engines) {
          engine = autocode.data.engines[engine_name];
          if (engine_name.match(new RegExp(value, 'i'))) {
            engines.push({
              action: {
                name: 'addExportEngine',
                data: { name: engine_name }
              },
              icon: 'https://cdn.rawgit.com/crystal/' + engine_name.split('.')[0] + '/master/.autocode/icon.svg',
              text: engine_name
            });
          }
        }
        
        autocode.fuzzy.open({
          rows: engines,
          target: $('#popup input[name="engine"]'),
          value: value
        });
      };
      
      autocode.popup.open({
        title: 'Edit Export Engine',
        content: form.toString()
      });
    },
    submit: function(data) {
      var data = {};
      $('#popup input, #popup select, #popup textarea').each(function() {
        data[$(this).attr('name')] = $(this).val();
      });
      
      autocode.project.exports[autocode.data.current.export].engine = data.engine;
      
      autocode.action.loadExport({ repo: data.name });
      
      autocode.popup.close();
      
      return false;
    }
  });
};;autocode.action.editExportFilename = function() {
  autocode.popup.open({
    title: 'Loading...'
  });
  autocode.popover.close();
  
  new formulator({
    data: {
      filename: autocode.project.exports[autocode.data.current.export].filename
    },
    formula: 'formulas/forms/Filename.json',
    xhr: true,
    ready: function(form) {
      autocode.popup.open({
        title: 'Edit Export Filename',
        content: form.toString()
      });
    },
    submit: function(data) {
      var data = {};
      $('#popup input, #popup select, #popup textarea').each(function() {
        data[$(this).attr('name')] = $(this).val();
      });
      
      autocode.project.exports[autocode.data.current.export].filename = data.filename;
      
      autocode.action.loadExport({ repo: data.name });
      
      autocode.popup.close();
      
      return false;
    }
  });
};;autocode.action.editExportFormat = function() {
  autocode.popup.open({
    title: 'Loading...'
  });
  autocode.popover.close();
  
  new formulator({
    data: {
      format: autocode.project.exports[autocode.data.current.export].format
    },
    formula: 'formulas/forms/Format.json',
    xhr: true,
    ready: function(form) {
      form.fields.format.autocomplete = false;
      form.fields.format.focus = function(o) {
        $(o).trigger('keyup');
      };
      form.fields.format.keydown = function() {
        if (event.keyCode == 13) {
          $('#fuzzy a').first().click();
        }
      };
      form.fields.format.keyup = function() {
        var rows = [], type_name, types = autocode.data.formats;
        for (var i in types) {
          type_name = types[i];
          if (type_name.match(new RegExp(value, 'i'))) {
            rows.push({
              action: {
                name: 'setExportFormat',
                data: { name: type_name }
              },
              icon: type_name.toLowerCase() + '-icon black',
              text: type_name
            });
          }
        }
        
        autocode.fuzzy.open({
          rows: rows,
          target: $('#popup input[name="format"]'),
          value: value
        });
      };
      
      autocode.popup.open({
        title: 'Edit Export Format',
        content: form.toString()
      });
    },
    submit: function(data) {
      var data = {};
      $('#popup input, #popup select, #popup textarea').each(function() {
        data[$(this).attr('name')] = $(this).val();
      });
      
      if (autocode.data.formats.indexOf(data.format.toLowerCase()) === -1) {
        autocode.popup.error('Unknown format: ' + data.format);
        return;
      }
      
      autocode.project.exports[autocode.data.current.export].format = data.format;
      
      autocode.action.loadExport({ repo: data.name });
      
      autocode.popup.close();
      
      return false;
    }
  });
};;autocode.action.editExportName = function() {
  autocode.popup.open({
    title: 'Loading...'
  });
  autocode.popover.close();
  
  new formulator({
    data: {
      name: autocode.data.current.export
    },
    formula: 'formulas/forms/Name.json',
    xhr: true,
    ready: function(form) {
      autocode.popup.open({
        title: 'Edit Export Name',
        content: form.toString()
      });
    },
    submit: function(data) {
      var data = {};
      $('#popup input, #popup select, #popup textarea').each(function() {
        data[$(this).attr('name')] = $(this).val();
      });
      
      if (data.name != autocode.data.current.export) {
        autocode.project.exports[data.name] = autocode.object.clone(autocode.project.exports[autocode.data.current.export]);
        delete(autocode.project.exports[autocode.data.current.export]);
      }
      autocode.project.exports = autocode.object.sort(autocode.project.exports);
      
      autocode.data.current.export = data.name;
      
      autocode.state['exports']();
      autocode.action.loadExport({ repo: data.name });
      
      autocode.popup.close();
      
      return false;
    }
  });
};;autocode.action.editExportSchema = function() {
  autocode.popup.open({
    title: 'Loading...'
  });
  autocode.popover.close();
  
  new formulator({
    data: {
      schema: autocode.project.exports[autocode.data.current.export].schema
    },
    formula: 'formulas/forms/Schema.json',
    xhr: true,
    ready: function(form) {
      form.fields.schema.autocomplete = false;
      form.fields.schema.focus = function(o) {
        $(o).trigger('keyup');
      };
      form.fields.schema.keyup = function() {
        var current_schemas = autocode.current.schemas();
        var schema, schema_icon, schema_style, schemas = [];
        for (var schema_name in current_schemas) {
          schema = current_schemas[schema_name];
          if (schema_name.match(new RegExp(value, 'i'))) {
            if (schema_name.split('.').length > 1) {
              schema_icon = 'https://cdn.rawgit.com/crystal/' + schema_name.split('.')[0] + '/master/.autocode/icon.svg';
              if (!schema_style && schema_style !== false) {
                schema_style = true;
              }
            } else {
              schema_icon = 'project-icon black';
            }
            schemas.push({
              action: {
                name: 'addExportSchema',
                data: { name: schema_name }
              },
              icon: schema_icon,
              style: schema_style ? 'divider' : null,
              text: schema_name
            });
            if (schema_style) {
              schema_style = false;
            }
          }
        }
        
        autocode.fuzzy.open({
          rows: schemas,
          target: $('#popup input[name="schema"]'),
          value: value
        });
      };
        
      autocode.popup.open({
        title: 'Edit Export Schema',
        content: form.toString()
      });
    },
    submit: function(data) {
      var data = {};
      $('#popup input, #popup select, #popup textarea').each(function() {
        data[$(this).attr('name')] = $(this).val();
      });
      
      autocode.project.exports[autocode.data.current.export].schema = data.schema;
      
      autocode.action.loadExport({ repo: data.name });
      
      autocode.popup.close();
      
      return false;
    }
  });
};;autocode.action.editExportType = function() {
  autocode.popup.open({
    title: 'Loading...'
  });
  autocode.popover.close();
  
  new formulator({
    data: {
      type: autocode.project.exports[autocode.data.current.export].type.substr(0, 1).toUpperCase()
        + autocode.project.exports[autocode.data.current.export].type.substr(1)
    },
    formula: 'formulas/forms/Type.json',
    xhr: true,
    ready: function(form) {
      form.fields.type.autocomplete = false;
      form.fields.type.focus = function(o) {
        $(o).trigger('keyup');
      };
      form.fields.type.keydown = function() {
        if (event.keyCode == 13) {
          $('#fuzzy a').first().click();
        }
      };
      form.fields.type.keyup = function() {
        var rows = [], type_name, types = autocode.data.exportTypes;
        for (var i in types) {
          type_name = types[i];
          if (type_name.match(new RegExp(value, 'i'))) {
            rows.push({
              action: {
                name: 'addExportName',
                data: { name: autocode.string.capitalize(type_name) }
              },
              icon: type_name.toLowerCase() + '-icon black',
              text: autocode.string.capitalize(type_name)
            });
          }
        }
        
        autocode.fuzzy.open({
          rows: rows,
          target: $('#popup input[name="type"]'),
          value: value
        });
      };
      
      autocode.popup.open({
        title: 'Edit Export Type',
        content: form.toString()
      });
    },
    submit: function(data) {
      var data = {};
      $('#popup input, #popup select, #popup textarea').each(function() {
        data[$(this).attr('name')] = $(this).val();
      });
      
      if (autocode.data.exportTypes.indexOf(data.type.toLowerCase()) === -1) {
        autocode.popup.error('Unknown type: ' + data.type);
        return;
      }
      
      autocode.project.exports[autocode.data.current.export].type = data.type.toLowerCase();
      
      autocode.action.loadExport({ repo: data.name });
      
      autocode.popup.close();
      
      return false;
    }
  });
};;autocode.action.editImportName = function() {
  autocode.popup.open({
    title: 'Loading...'
  });
  autocode.popover.close();
  
  new formulator({
    data: {
      name: autocode.data.current.import
    },
    formula: 'formulas/forms/Name.json',
    xhr: true,
    ready: function(form) {
      autocode.popup.open({
        title: 'Edit Import Name',
        content: form.toString()
      });
    },
    submit: function(data) {
      var data = {};
      $('#popup input, #popup select, #popup textarea').each(function() {
        data[$(this).attr('name')] = $(this).val();
      });
      
      if (data.name != autocode.data.current.import) {
        autocode.project.imports[data.name] = autocode.object.clone(autocode.project.imports[autocode.data.current.import]);
        delete(autocode.project.imports[autocode.data.current.import]);
      }
      
      autocode.popup.close();
      
      autocode.state['imports']({ disableSelected: true });
      autocode.action.loadImport({ repo: data.name });
      
      return false;
    }
  });
};;autocode.action.editImportVersion = function() {
  autocode.popup.open({
    title: 'Loading...'
  });
  autocode.popover.close();
  
  new formulator({
    data: {
      version: autocode.project.imports[autocode.data.current.import]
    },
    formula: 'formulas/forms/Version.json',
    xhr: true,
    ready: function(form) {
      autocode.popup.open({
        title: 'Edit Import Name',
        content: form.toString()
      });
    },
    submit: function(data) {
      var data = {};
      $('#popup input, #popup select, #popup textarea').each(function() {
        data[$(this).attr('name')] = $(this).val();
      });
      
      if (!data.version || !data.version.length) {
        return autocode.popup.error('Version is required.');
      }
      
      autocode.project.imports[autocode.data.current.import] = data.version;
      
      autocode.popup.close();
      
      autocode.state['imports']({ disableSelected: true });
      autocode.action.loadImport({ repo: autocode.data.current.import });
      
      return false;
    }
  });
};;autocode.action.editInterfaceName = function() {
  autocode.popup.open({
    title: 'Loading...'
  });
  autocode.popover.close();
  
  new formulator({
    data: autocode.project.interfaces[autocode.data.current.interface],
    formula: 'formulas/forms/InterfaceName.json',
    xhr: true,
    ready: function(form) {
      autocode.popup.open({
        title: 'Edit Interface Name',
        content: form.toString()
      });
    },
    submit: function(data) {
      var data = {};
      $('#popup input, #popup select, #popup textarea').each(function() {
        data[$(this).attr('name')] = $(this).val();
      });
      
      if (data.name && data.name.length) {
        autocode.project.interfaces[autocode.data.current.interface].name = data.name;
      } else {
        delete(autocode.project.interfaces[autocode.data.current.interface].name);
      }
      
      autocode.popup.close();
      
      autocode.state['interfaces']();
      autocode.action.loadInterface();
      
      return false;
    }
  });
};;autocode.action.editInterfacePort = function() {
  autocode.popup.open({
    title: 'Loading...'
  });
  autocode.popover.close();
  
  new formulator({
    data: autocode.project.interfaces[autocode.data.current.interface],
    formula: 'formulas/forms/InterfacePort.json',
    xhr: true,
    ready: function(form) {
      autocode.popup.open({
        title: 'Edit Interface Port',
        content: form.toString()
      });
    },
    submit: function(data) {
      var data = {};
      $('#popup input, #popup select, #popup textarea').each(function() {
        data[$(this).attr('name')] = $(this).val();
      });
      
      if (data.port && data.port.length) {
        autocode.project.interfaces[autocode.data.current.interface].port = data.port;
      } else {
        delete(autocode.project.interfaces[autocode.data.current.interface].port);
      }
      
      autocode.popup.close();
      
      autocode.state['interfaces']();
      autocode.action.loadInterface();
      
      return false;
    }
  });
};;autocode.action.editName = function() {
  autocode.popup.open({
    title: 'Loading...'
  });
  autocode.popover.close();
  
  new formulator({
    data: autocode.project,
    formula: 'formulas/forms/Name.json',
    xhr: true,
    ready: function(form) {
      autocode.popup.open({
        title: 'Edit Project Name',
        content: form.toString()
      });
    },
    submit: function(data) {
      var data = {};
      $('#popup input, #popup select, #popup textarea').each(function() {
        data[$(this).attr('name')] = $(this).val();
      });
      
      if (data.name && data.name.length) {
        autocode.project.name = data.name;
      } else {
        delete(autocode.project.name);
      }
      
      autocode.popup.close();
      
      autocode.state['overview/general']();
      
      return false;
    }
  });
};;autocode.action.editOutputFilename = function() {
  autocode.popup.open({
    title: 'Loading...'
  });
  autocode.popover.close();
  
  new formulator({
    data: {
      filename: autocode.project.outputs[autocode.data.current.output].filename
    },
    formula: 'formulas/forms/Filename.json',
    xhr: true,
    ready: function(form) {
      autocode.popup.open({
        title: 'Edit Output Filename',
        content: form.toString()
      });
    },
    submit: function(data) {
      var data = {};
      $('#popup input, #popup select, #popup textarea').each(function() {
        data[$(this).attr('name')] = $(this).val();
      });
      
      autocode.project.outputs[autocode.data.current.output].filename = data.filename;
      
      autocode.state['outputs']();
      autocode.action.loadOutput();
      
      autocode.popup.close();
      
      return false;
    }
  });
};;autocode.action.editOutputGenerator = function() {
  autocode.popup.open({
    title: 'Loading...'
  });
  autocode.popover.close();
  
  new formulator({
    data: {
      generator: autocode.project.outputs[autocode.data.current.output].generator
    },
    formula: 'formulas/forms/Generator.json',
    xhr: true,
    ready: function(form) {
      form.fields.generator.autocomplete = false;
      form.fields.generator.focus = function(o) {
        $(o).trigger('keyup');
      };
      form.fields.generator.keydown = function() {
        if (event.keyCode == 13) {
          $('#fuzzy a').first().click();
        }
      };
      form.fields.generator.keyup = function() {
        var value = $('#popup input[name="generator"]').val();
        
        var current_generators = autocode.current.generators();
        var generator, generator_icon, generators = [];
        for (var generator_name in current_generators) {
          generator = current_generators[generator_name];
          if (generator_name.split('.').length > 1) {
            generator_icon = 'https://cdn.rawgit.com/crystal/' + generator_name.split('.')[0] + '/master/.autocode/icon.svg';
          } else {
            generator_icon = 'project-icon black';
          }
          if (generator_name.match(new RegExp(value, 'i'))) {
            generators.push({
              action: {
                name: 'addOutputGenerator',
                data: {
                  name: generator_name
                }
              },
              icon: generator_icon,
              text: generator_name
            });
          }
        }
        autocode.fuzzy.open({
          rows: generators,
          target: $('#popup input[name="generator"]'),
          value: value
        });
        
        $('#popup input[name="filename"]').attr('placeholder', current_generators[value] && current_generators[value].filename ? current_generators[value].filename : '');
      };
      
      autocode.popup.open({
        title: 'Edit Output Generator',
        content: form.toString()
      });
    },
    submit: function(data) {
      var data = {};
      $('#popup input, #popup select, #popup textarea').each(function() {
        data[$(this).attr('name')] = $(this).val();
      });
      
      autocode.project.outputs[autocode.data.current.output].generator = data.generator;
      
      autocode.state['outputs']();
      autocode.action.loadOutput();
      
      autocode.popup.close();
      
      return false;
    }
  });
};;autocode.action.editOutputPath = function() {
  autocode.popup.open({
    title: 'Loading...'
  });
  autocode.popover.close();
  
  new formulator({
    data: {
      path: autocode.project.outputs[autocode.data.current.output].path
    },
    formula: 'formulas/forms/Path.json',
    xhr: true,
    ready: function(form) {
      autocode.popup.open({
        title: 'Edit Output Path',
        content: form.toString()
      });
    },
    submit: function(data) {
      var data = {};
      $('#popup input, #popup select, #popup textarea').each(function() {
        data[$(this).attr('name')] = $(this).val();
      });
      
      autocode.project.outputs[autocode.data.current.output].path = data.path;
      
      autocode.state['outputs']();
      autocode.action.loadOutput();
      
      autocode.popup.close();
      
      return false;
    }
  });
};;autocode.action.editProjectCopyright = function() {
  autocode.popup.open({
    title: 'Loading...'
  });
  autocode.popover.close();
  
  new formulator({
    data: {
      copyright: autocode.project.copyright
    },
    formula: 'formulas/forms/Copyright.json',
    xhr: true,
    ready: function(form) {
      autocode.popup.open({
        title: 'Edit Project Copyright',
        content: form.toString()
      });
    },
    submit: function(data) {
      var data = {};
      $('#popup input, #popup select, #popup textarea').each(function() {
        data[$(this).attr('name')] = $(this).val();
      });
      
      if (data.copyright && data.copyright.length) {
        autocode.project.copyright = data.copyright;
      } else {
        delete(autocode.project.copyright);
      }
      
      autocode.popup.close();
      
      autocode.state['overview/general']();
      
      return false;
    }
  });
};;autocode.action.editURL = function() {
  autocode.popup.open({
    title: 'Loading...'
  });
  autocode.popover.close();
  
  new formulator({
    data: autocode.project,
    formula: 'formulas/forms/URL.json',
    xhr: true,
    ready: function(form) {
      autocode.popup.open({
        title: 'Edit Project URL',
        content: form.toString()
      });
    },
    submit: function(data) {
      var data = {};
      $('#popup input, #popup select, #popup textarea').each(function() {
        data[$(this).attr('name')] = $(this).val();
      });
      
      if (data.url && data.url.length) {
        autocode.project.url = data.url;
      } else {
        delete(autocode.project.url);
      }
      
      autocode.popup.close();
      
      autocode.state['overview/general']();
      
      return false;
    }
  });
};;autocode.action.editVersion = function() {
  autocode.popup.open({
    title: 'Loading...'
  });
  autocode.popover.close();
  
  new formulator({
    data: autocode.project,
    formula: 'formulas/forms/Version.json',
    xhr: true,
    ready: function(form) {
      autocode.popup.open({
        title: 'Edit Project Version',
        content: form.toString()
      });
    },
    submit: function(data) {
      var data = {};
      $('#popup input, #popup select, #popup textarea').each(function() {
        data[$(this).attr('name')] = $(this).val();
      });
      
      if (data.version && data.version.length) {
        autocode.project.version = data.version;
      } else {
        delete(autocode.project.version);
      }
      
      autocode.popup.close();
      
      autocode.state['overview/general']();
      
      return false;
    }
  });
};;autocode.action.editorSettings = function() {
  autocode.popup.open({
    title: 'Loading...'
  });
  autocode.popover.close();
  
  new formulator({
    data: {
      lineNumbers: autocode.editor.lineNumbersEnabled(),
      theme: autocode.editor.getTheme()
    },
    formula: 'formulas/forms/Editor.json',
    xhr: true,
    ready: function(form) {
      form.fields.theme.autocomplete = false;
      form.fields.theme.focus = function(o) {
        $(o).trigger('keyup');
      };
      form.fields.theme.keyup = function() {
        var rows = [], type_name, types = autocode.editor.codemirror.themes;
        for (var i in types) {
          type_name = types[i];
          if (type_name.match(new RegExp(value, 'i'))) {
            rows.push({
              action: {
                name: 'setEditorTheme',
                data: { name: autocode.string.capitalize(type_name) }
              },
              text: autocode.string.capitalize(type_name)
            });
          }
        }
        
        autocode.fuzzy.open({
          rows: rows,
          target: $('#popup input[name="theme"]'),
          value: value
        });
      };
      
      autocode.popup.open({
        title: 'Editor Settings',
        content: '<textarea style="display: none"></textarea>' + form.toString(),
        complete: function() {
          var editor = CodeMirror.fromTextArea($('#popup textarea')[0], {
            lineNumbers: true,
            mode: 'javascript',
            readOnly: true,
            theme: autocode.editor.getTheme()
          });
          
          code_mirror = $('#popup .CodeMirror');
          code_mirror[0].CodeMirror.setValue("function findSequence(goal) {\n  function find(start, history) {\n    if (start == goal)\n      return history;\n    else if (start > goal)\n      return null;\n    else\n      return find(start + 5, \"(\" + history + \" + 5)\") ||\n             find(start * 3, \"(\" + history + \" * 3)\");\n  }\n  return find(1, \"1\");\n}");
          
          autocode.resize.popup();
        }
      });
    },
    submit: function(data) {
      var data = {};
      $('#popup input, #popup select, #popup textarea').each(function() {
        if (!$(this).attr('name') || !$(this).attr('name').length) {
          return;
        }
        if ($(this).attr('type') == 'checkbox') {
          if ($(this).is(':checked')) {
            data[$(this).attr('name')] = true;
          } else {
            data[$(this).attr('name')] = false;
          }
        } else {
          data[$(this).attr('name')] = $(this).val();
        }
      });
      
      data.theme = data.theme.toLowerCase();
      
      if (autocode.editor.codemirror.themes.indexOf(data.theme) === -1) {
        autocode.popup.error('Unknown theme: ' + data.theme);
        return;
      }
      
      console.log(data);
      
      if (data.lineNumbers) {
        autocode.editor.enableLineNumbers();
      } else {
        autocode.editor.disableLineNumbers();
      }
      autocode.editor.setTheme(data.theme);
      
      autocode.fuzzy.close();
      autocode.popup.close();
      
      return false;
    }
  });
};;autocode.action.generate = function() {
  if (autocode.data.current.tab == 'config') {
    var config = $('#config-content .CodeMirror')[0].CodeMirror.getValue();
    autocode.project = jsyaml.safeLoad(config);
  }
  
  if (!autocode.project.outputs || !autocode.project.imports) {
    if (!autocode.project.imports) {
      autocode.popup.open({
        title: 'Unable to Generate Project',
        content: '<div style="padding-bottom: 10px">You haven\'t imported any modules yet.</div>'
          + '<a class="button" href="imports">Import a Module</a> <button class="secondary" onclick="autocode.action.closePopup()">Cancel</button>'
      });
    } else {
      autocode.popup.open({
        title: 'Unable to Generate Project',
        content: '<div style="padding-bottom: 10px">You haven\'t added any outputs to your Autocode configuration.</div>'
          + '<a class="button" href="config">Add an Output</a> <button class="secondary" onclick="autocode.action.closePopup()">Cancel</button>'
      });
    }
    
    return;
  }
  
  autocode.ws.io.emit('generate', {
    config: autocode.project,
    project: autocode.repo.split('/')[1],
    user: autocode.repo.split('/')[0]
  });
};;autocode.action.githubRepo = function() {
  autocode.popover.close();
  
  window.open('https://github.com/' + autocode.repo);
};;autocode.action.githubUser = function() {
  autocode.popover.close();
  
  window.open('https://github.com/' + autocode.data.user.username);
};;autocode.action.halt = function() {
  // close popover
  autocode.popover.close();
  
  // halt process
  autocode.ws.io.emit('halt');
};;autocode.action.import = function() {
  autocode.popup.open({
    title: 'Loading Modules...'
  });
  
  autocode.api.modules.get({
    success: function(modules) {
      var rows = [];
      for (var i = 0; i < modules.length; i++) {
        rows.push({
          action: {
            name: 'addImport',
            data: {
              repo: modules[i].name
            }
          },
          icon: 'https://cdn.rawgit.com/' + modules[i].name + '/master/.autocode/icon.svg',
          text: modules[i].name
        });
      }
      
      autocode.popup.open({
        title: 'Import Modules',
        rows: rows,
        style: 'table'
      });
    }
  });
};;autocode.action.importsExportsTab = function(opts) {
  opts = opts || {};
  
  $('#imports-exports .imports-export').hide();
  $('#imports-exports-' + opts.name).show();
};;autocode.action.importsTab = function(opts) {
  opts = opts || {};
  
  $('#imports-exports .imports-export').hide();
  $('#imports-exports-' + opts.name).show();
};;autocode.action.initProject = function() {
  autocode.popup.open({
    title: 'Initializing Autocode...'
  });
  autocode.popover.close();
  
  autocode.api.init.post({
    data: {
      repo: autocode.repo
    },
    success: function(data) {
      autocode.action.loadProject({ name: autocode.repo });
    }
  });
};;autocode.action.launchInterface = function() {
  var state = 'http://' + autocode.ws.ip;
  var interface_data = autocode.project.interfaces[autocode.data.current.interface];
  var port = interface_data.port;
  var path = interface_data.path;
  var params = interface_data.params;
  var name = interface_data.name;
  
  if (port) {
    if (typeof(port) == 'string' && port.substr(0, 1) == '$' && autocode.project[port.substr(1)]) {
      port = autocode.project[port.substr(1)];
    }
    state += ':' + port;
  }
  if (path) {
    if (typeof(path) == 'string' && path.substr(0, 1) == '$' && autocode.project[path.substr(1)]) {
      path = autocode.project[path.substr(1)];
    }
    state += path;
  }
  if (params) {
    if (typeof(params) == 'string' && params.substr(0, 1) == '$' && autocode.project[params.substr(1)]) {
      params = autocode.project[params.substr(1)];
    }
    state += '?' + autocode.query.build(params);
  }
  if (typeof(name) == 'string' && name.substr(0, 1) == '$' && autocode.project[name.substr(1)]) {
    name = autocode.project[name.substr(1)];
  }
  
  window.open(state, '_blank');
};;autocode.action.loadExport = function(opts) {
  opts = opts || {};
  
  autocode.data.current.export = opts.name || autocode.data.current.export || autocode.object.firstKey(autocode.project.exports);
  
  var export_data = autocode.project.exports[autocode.data.current.export];
  
  $('#exports-content-container .table a').removeClass('selected');
  $('#exports-' + autocode.data.current.export).addClass('selected');
  
  $('#exports-name .value').text(autocode.data.current.export || ' [ Click to Add ]');
  $('#exports-type .value').text(export_data.type.substr(0, 1).toUpperCase() + export_data.type.substr(1) || ' [ Click to Add ]');
  $('#exports-description .value').html(export_data.description ? marked(export_data.description) : ' [ Click to Add ]');
  $('#exports-engine .value').text(export_data.engine || ' [ Click to Add ]');
  $('#exports-filename .value').text(export_data.filename || ' [ Click to Add ]');
  $('#exports-format .value').text(export_data.format || ' [ Click to Add ]');
  $('#exports-schema .value').text(export_data.schema || ' [ Click to Add ]');
  
  var code_mirror = $('#exports-content .CodeMirror');
  var mode = 'text', tabs = [], value = jsyaml.safeDump(autocode.project);
  switch (export_data.type) {
    case 'engine': {
      $('#exports-engine').hide();
      $('#exports-filename').hide();
      $('#exports-format').hide();
      $('#exports-schema').hide();
      
      tabs = ['engine'];
      
      mode = 'javascript';
      value = export_data.engine || "\n";
      
      break;
    }
    case 'generator': {
      $('#exports-engine').show();
      $('#exports-filename').show();
      $('#exports-format').show();
      $('#exports-schema').show();
      
      tabs = ['template'];
      
      mode = export_data.format;
      value = export_data.template || "\n";
      
      break;
    }
    case 'helper': {
      $('#exports-engine').hide();
      $('#exports-filename').hide();
      $('#exports-format').hide();
      $('#exports-schema').hide();
      
      tabs = ['helper'];
      
      mode = 'javascript';
      value = export_data.helper || "\n";
      
      break;
    }
    case 'processor': {
      $('#exports-engine').hide();
      $('#exports-filename').hide();
      $('#exports-format').hide();
      $('#exports-schema').hide();
      
      tabs = ['processor'];
      
      mode = 'javascript';
      value = export_data.processor || "\n";
      
      break;
    }
    case 'schema': {
      $('#exports-engine').hide();
      $('#exports-filename').hide();
      $('#exports-format').hide();
      $('#exports-schema').hide();
      
      tabs = ['schema'];
      
      mode = 'yaml';
      value = export_data.schema ? jsyaml.safeDump(export_data.schema) : "\n";
      
      break;
    }
    case 'spec': {
      $('#exports-engine').hide();
      $('#exports-filename').hide();
      $('#exports-format').hide();
      $('#exports-schema').hide();
      
      tabs = ['spec'];
      
      mode = 'yaml';
      value = export_data.spec ? jsyaml.safeDump(export_data.spec) : "\n";
      
      break;
    }
    case 'transformer': {
      $('#exports-engine').hide();
      $('#exports-filename').hide();
      $('#exports-format').hide();
      $('#exports-schema').hide();
      
      tabs = ['transformer'];
      
      mode = 'javascript';
      value = export_data.transformer || "\n";
      
      break;
    }
  }
  $('#exports-tabs').text('');
  
  var tab;
  for (var i = 0; i < tabs.length; i++) {
    tab = tabs[i];
    $('#exports-tabs').append('<a' + (i == 0 ? ' class="selected"' : '') + ' id="exports-' + tab + '-tab" onclick="autocode.action.exportsTab({ tab: \'' + tab + '\' })">' + tab.substr(0,1).toUpperCase() + tab.substr(1) + '</a>')
  }
  
  if (!code_mirror.length) {
    var editor = CodeMirror.fromTextArea($('#exports-content textarea')[0], {
      lineNumbers: autocode.editor.lineNumbersEnabled(),
      mode: mode
    });
    
    code_mirror = $('#exports-content .CodeMirror')
    code_mirror[0].CodeMirror.setValue("\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n");
    code_mirror[0].CodeMirror.setValue(value);
    
    $('.CodeMirror-scroll').scrollTop(2);
    
  } else {
    code_mirror[0].CodeMirror.setValue("\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n");
    code_mirror[0].CodeMirror.setValue(value);
  }
  
  code_mirror[0].CodeMirror.setOption('mode', mode);
  code_mirror[0].CodeMirror.setOption('theme', autocode.editor.getTheme());
  
  setTimeout(function() { $('#exports-content .CodeMirror')[0].CodeMirror.refresh() }, 0);
  
  autocode.hint.init();
  
  autocode.resize.all();
};;autocode.action.loadFile = function(opts) {
  autocode.action.toggleColumn('output-content', 2);
  
  var code_mirror = $('#output-content .CodeMirror');
  var value = opts.value;
  
  var mode = 'text';
  switch (true) {
    case !!opts.file.match(/\.coffee$/): {
      mode = 'coffeescript';
      break;
    }
    case opts.file == 'Dockerfile': {
      mode = 'dockerfile';
      break;
    }
    case !!opts.file.match(/\.htm$/) || !!opts.file.match(/\.html$/): {
      mode = 'htmlmixed';
      break;
    }
    case !!opts.file.match(/\.ini$/): {
      mode = 'ini';
      break;
    }
    case opts.file == '.bowerrc'
      || opts.file == '.eslintrc'
      || !!opts.file.match(/\.js$/)
      || !!opts.file.match(/\.json$/): {
      mode = 'javascript';
      break;
    }
    case opts.file == 'AUTHORS' || !!opts.file.match(/\.md$/): {
      mode = 'markdown';
      break;
    }
    case !!opts.file.match(/\.py$/): {
      mode = 'python';
      break;
    }
    case !!opts.file.match(/\.rb$/): {
      mode = 'ruby';
      break;
    }
    case !!opts.file.match(/\.sass$/): {
      mode = 'sass';
      break;
    }
    case !!opts.file.match(/\.yml$/) || !!opts.file.match(/\.yaml$/): {
      mode = 'yaml';
      break;
    }
  }
  
  if (!code_mirror.length) {
    var editor = CodeMirror.fromTextArea($('#output-content textarea')[0], {
      lineNumbers: autocode.editor.lineNumbersEnabled(),
      readOnly: true,
      mode: mode,
      theme: autocode.editor.getTheme()
    });
    
    code_mirror = $('#output-content .CodeMirror')
    code_mirror[0].CodeMirror.setValue(value);
    
    $('.CodeMirror-scroll').scrollTop(2);
    
  } else {
    code_mirror[0].CodeMirror.setOption('mode', mode);
    code_mirror[0].CodeMirror.setValue(value);
  }
  
  setTimeout(function() { $('#output-content .CodeMirror')[0].CodeMirror.refresh() }, 0);
  
  autocode.resize.all();
};;autocode.action.loadImport = function(opts) {
  autocode.action.toggleColumn('imports-content', 2);
  
  autocode.data.current.import = opts.repo;
  
  $('#imports-content-container .table a').removeClass('selected');
  $('#imports-content-container .table a').eq(opts.index).addClass('selected');
  
  $('#imports-name .value').text(opts.repo);
  $('#imports-version .value').text(autocode.project.imports[opts.repo]);
  $('#imports-content-readme').text('');
  
  $('#imports-tabs, #imports-exports .imports-export').hide();
  
  autocode.loader.open();
  
  autocode.api.config.get({
    data: {
      repo: opts.repo
    },
    success: function(data) {
      // close loader
      autocode.loader.close();
      
      // load config
      var config = jsyaml.safeLoad(data.config);
      
      if (autocode.project.imports) {
        $('#imports-content-container .content-right').show();
      } else {
        $('#imports-content-container .content-right').hide();
      }
      
      $('#imports-tabs, #imports-exports .imports-export').text('').show();
      
      var tab, tabs = [], export_type;
      for (var export_name in config.exports) {
        export_type = config.exports[export_name].type + 's';
        if (tabs.indexOf(export_type) === -1) {
          tabs.push(export_type);
        }
      }
      for (var i = 0; i < tabs.length; i++) {
        tab = tabs[i];
        $('#imports-tabs').append('<a' + (i == 0 ? ' class="selected"' : '') + ' id="imports-' + tab + '-tab" onclick="autocode.action.importsTab({ name: \'' + tab + '\' })">' + tab.substr(0,1).toUpperCase() + tab.substr(1) + '</a>');
      }
      
      for (var export_name in config.exports) {
        $('#imports-exports-' + config.exports[export_name].type + 's').append('<h2>' + export_name + '</h2><p>' + (config.exports[export_name].description ? marked(config.exports[export_name].description) : '<em>There is no description for this ' + config.exports[export_name].type + '.</em>') + '</p>');
      }
      $('#imports-exports .imports-export').hide();
      
      $('#imports-tabs a').first().click();
      
      autocode.resize.content();
    }
  });
};;autocode.action.loadInterface = function(opts) {
  opts = opts || {};
  
  autocode.data.current.interface = opts.index || autocode.data.current.interface;
  
  $('#interfaces-content-container .table a').removeClass('selected');
  $('#interfaces-' + autocode.data.current.interface).addClass('selected');
  
  var interface_name = autocode.project.interfaces[autocode.data.current.interface].name;
  if (typeof(autocode.project.interfaces[autocode.data.current.interface].name) == 'string' && autocode.project.interfaces[autocode.data.current.interface].name.substr(0, 1) == '$' && autocode.project[autocode.project.interfaces[autocode.data.current.interface].name.substr(1)]) {
    interface_name = autocode.project[autocode.project.interfaces[autocode.data.current.interface].name.substr(1)];
  }
  
  $('#interfaces-title').text(interface_name);
  
  autocode.hint.init();
  
  autocode.resize.all();
};;autocode.action.loadOutput = function(opts) {
  opts = opts || {};
  autocode.data.current.output = opts.output || autocode.data.current.output;
  
  var output = autocode.project.outputs[autocode.data.current.output];
  var generator = autocode.project.exports && autocode.project.exports[output.generator]
    ? autocode.project.exports[output.generator]
    : null;
  var schema;
  if (generator) {
    if (typeof(generator.schema) == 'string') {
      schema = autocode.project.exports[generator.schema].schema;
    } else {
      schema = generator.schema;
    }
  } else if (!generator && autocode.imports[output.generator.split('.')[0]]) {
    var imported = autocode.imports[output.generator.split('.')[0]];
    if (imported && imported.exports && imported.exports[output.generator.split('.')[1]]) {
      generator = imported.exports[output.generator.split('.')[1]];
      if (typeof(generator.schema) == 'string') {
        schema = imported.exports[generator.schema].schema;
      } else {
        schema = generator.schema;
      }
    }
  }
  
  autocode.data.current.generator = output.generator;
  
  var filename = output.filename;
  if (!filename && generator.filename) {
    filename = '[ ' + generator.filename + ' ]';
  } else if (!filename) {
    filename = '[ Click to Add ]';
  }
  
  $('#outputs-content-container .table a').removeClass('selected');
  $('#outputs-' + autocode.data.current.output).addClass('selected');
  
  $('#outputs-filename .value').text(filename);
  $('#outputs-generator .value').text(output.generator);
  $('#outputs-path .value').text(output.path || '[ Project Root ]');
  
  if (output.spec) {
    $('#outputs-content textarea').val(jsyaml.safeDump(output.spec));
  } else {
    $('#outputs-content textarea').val('');
  }
  
  $('#outputs-content .content-center .schema').empty();
  
  autocode.data.autocomplete = [];
  if (schema) {
    function loadSchemaValues(properties, tab) {
      var schema_property;
      for (var schema_name in properties) {
        if (autocode.data.autocomplete.indexOf(schema_name) === -1) {
          autocode.data.autocomplete.push(tab + schema_name);
        }
        schema_property = properties[schema_name];
        if (schema_property.properties) {
          loadSchemaValues(schema_property.properties, schema_name + '.');
        }
      }
    };
    loadSchemaValues(schema.properties, '');
  }
  for (var global_var in autocode.project) {
    autocode.data.autocomplete.push('$' + global_var);
  }
  
  var code_mirror = $('#outputs-content .CodeMirror');
  var mode = 'yaml';
  var value = autocode.project.outputs[autocode.data.current.output].spec
    ? jsyaml.safeDump(autocode.project.outputs[autocode.data.current.output].spec)
    : '';
  if (!code_mirror.length) {
    var editor = CodeMirror.fromTextArea($('#outputs-content textarea')[0], {
      extraKeys: {
        "Ctrl-Space": 'autocomplete',
        Tab: function(cm) {
          if (cm.somethingSelected()) {
            cm.indentSelection("add");
            return;
          }
          var spaces = Array(cm.getOption("indentUnit") + 1).join(" ");
          cm.replaceSelection(spaces);
        },
        'Shift-Tab': function(cm) {
          cm.indentSelection("subtract");
        }
      },
      lineNumbers: autocode.editor.lineNumbersEnabled(),
      mode: mode,
      theme: autocode.editor.getTheme()
    });
    
    code_mirror = $('#outputs-content .CodeMirror')
    code_mirror[0].CodeMirror.on('keyup', function(cm, event) {
      if (!cm.state.completionActive && event.keyCode != 13 && event.keyCode != 27) {
        CodeMirror.commands.autocomplete(cm, null, {completeSingle: false});
      }
    });
    code_mirror[0].CodeMirror.setValue("\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n");
    code_mirror[0].CodeMirror.setValue(value);
    
    $('.CodeMirror-scroll').scrollTop(2);
    
  } else {
    code_mirror[0].CodeMirror.setValue("\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n");
    code_mirror[0].CodeMirror.setValue(value);
  }
  
  setTimeout(function() { $('#outputs-content .CodeMirror')[0].CodeMirror.refresh() }, 0);
  
  autocode.action.outputsHide();
  
  autocode.hint.init();
  
  // prepare forms
  
  /*
  
  var schema = generator.schema;
  if (typeof(schema) == 'string') {
    schema = schema.split('.');
    if (schema.length > 1) {
      var schema_prefix = schema[0];
      var schema_suffix = schema[1];
      if (!autocode.imports[schema_prefix] || !autocode.imports[schema_prefix].exports[schema_suffix].schema) {
        return;
      }
      schema = autocode.imports[schema_prefix].exports[schema_suffix].schema;
    } else {
      schema = schema[0];
      if (!autocode.project.exports[schema] || !autocode.project.exports[schema].schema) {
        return;
      }
      schema = autocode.project.exports[schema].schema;
    }
  }
  
  if (!schema || !schema.properties) {
    return;
  }
  
  function deepFind(obj, path) {
    var paths = path.split('.')
      , current = obj
      , i;

    for (i = 0; i < paths.length; ++i) {
      if (!current || current[paths[i]] == undefined) {
        return '';
      } else {
        current = current[paths[i]];
      }
    }
    return current;
  };
  
  function addProperties(properties, parent) {
    var property,
      property_add,
      property_field,
      property_icon,
      property_input,
      property_label,
      property_name,
      property_text,
      property_title,
      property_variable;
    for (var property_name in autocode.object.sort(properties)) {
      property = properties[property_name];
      
      if (property.title) {
        property_title = property.title;
      } else {
        property_title = property_name;
      }
      
      if (property.type == 'object' && !property.additionalProperties) {
        $('#outputs-content .content-center .schema').append('<b style="display: block; padding-top: 10px">' + property_title + '</b>');
        addProperties(property.properties, property_name);
        continue;
      }
      
      property_field = $(document.createElement('div'));
      if (parent) {
        property_field.css('paddingLeft', 20);
      }
      property_field.addClass('field');
      
      property_label = $(document.createElement('label'));
      property_field.append(property_label);
      
      property_text = $(document.createElement('span'));
      property_text.addClass('text');
      property_text.text(property_title);
      property_label.append(property_text);
      
      if (property.description) {
        property_hint = autocode.string.escape(
          marked(property.description) + (property.default || property.default === false ? '<div style="font-weight: bold">Default: ' + property.default + '</div>' : '')
        );
        property_icon = $(document.createElement('span'));
        property_icon.addClass('icon info-icon');
        property_icon.attr('data-hint', property_hint);
        property_label.append(property_icon);
      }
      
      if (property.type == 'object') {
        property_add = $(document.createElement('button'));
        property_add.attr('type', 'button');
        property_add.data('name', property_name);
        property_add.click(function() {
          var i = ($(this).parent().parent().children().length - 1) / 2;
          
          var input = autocode.element.input.html({
            autocomplete: false,
            class: property_variable ? 'variable' : null,
            css: {
              borderRight: 'none',
              width: '50%'
            },
            event: {
              blur: autocode.state['outputs/input/blur'],
              focus: autocode.state['outputs/input/focus'],
              keydown: function(e) {
                if (e.keyCode == 27) {
                  $(this).blur();
                  return;
                }
              }
            },
            name: $(this).data('name') + '[0][key]',
            placeholder: property.default + '(' + property.type + ')',
            spellcheck: false,
            type: 'text',
            value: property_value
          });
          $(this).parent().parent().append(input);
          
          var input = autocode.element.input.html({
            autocomplete: false,
            class: property_variable ? 'variable' : null,
            css: {
              width: '50%'
            },
            event: {
              blur: autocode.state['outputs/input/blur'],
              focus: autocode.state['outputs/input/focus'],
              keydown: function(e) {
                if (e.keyCode == 27) {
                  $(this).blur();
                  return;
                }
              }
            },
            name: $(this).data('name') + '[0][value]',
            placeholder: property.default,
            spellcheck: false,
            type: 'text',
            value: property_value
          });
          $(this).parent().parent().append(input);
        });
        property_add.text('Add');
        property_label.append(property_add);
        
        property_add.click();
        
      } else if (property.type == 'boolean') {
        property_value = (output.spec && (output.spec[property_name] || output.spec[property_name] === false) ? output.spec[property_name] : null);
        property_input = autocode.element.radio.html({
          defaultValue: property.default,
          name: property_name,
          options: {
            true: 'True',
            false: 'False'
          },
          value: property_value
        });
        property_field.append(property_input);
        
      } else {
        property_value = deepFind(output.spec, parent ? parent + '.' + property_name : property_name);
        if (property_value.substr(0, 1) == '$' && autocode.project[property_value.substr(1)]) {
          property_variable = true;
          property_value = autocode.project[property_value.substr(1)];
        } else {
          property_variable = false;
        }
        property_input = autocode.element.input.html({
          autocomplete: false,
          class: property_variable ? 'variable' : null,
          event: {
            blur: autocode.state['outputs/input/blur'],
            focus: autocode.state['outputs/input/focus'],
            keydown: function(e) {
              if (e.keyCode == 27) {
                $(this).blur();
                return;
              }
            }
          },
          name: property_name,
          placeholder: property.default,
          spellcheck: false,
          type: 'text',
          value: property_value
        });
        property_field.append(property_input);
      }
      
      $('#outputs-content .content-center .schema').append(property_field);
    }
  };
  
  addProperties(schema.properties);
  */
};;autocode.action.loadProject = function(opts) {
  opts = opts || {};
  
  autocode.popover.close();
  
  // ask to close project before loading another
  if (opts.force !== true && autocode.data.originalConfig && autocode.project && autocode.data.originalConfig != jsyaml.safeDump(autocode.project)) {
    autocode.popup.open({
      title: 'Close Project',
      content: '<div style="padding-bottom: 15px">Are you sure you want to close this project and load another? <b>You will lose all unsaved changes.</b></div>'
        + '<button onclick="autocode.action.closePopup()">No, Keep It Open</button> <a class="button secondary" onclick="autocode.action.loadProject({ force: true })">Yes, Close Project</a> <a class="button secondary" href="project/diff">View Unsaved Changes</a>'
    });
    return;
  }
  
  // get projects to load
  if (!opts.name) {
    autocode.popup.open({
      title: 'Loading...'
    });
    
    autocode.api.repos.get({
      success: function(data) {
        autocode.data.projects = data;
        
        new formulator({
          formula: 'formulas/forms/Name.json',
          xhr: true,
          ready: function(form) {
            form.fields.name.autocomplete = false;
            
            form.fields.name.keydown = function() {
              if (event.keyCode == 13) {
                $('#fuzzy a').first().click();
                return false;
              }
            };
            form.fields.name.keyup = function() {
              var value = $('#popup input[name="name"]').val();
              
              var project, projects = [];
              for (var project_i in autocode.data.projects) {
                project = autocode.data.projects[project_i];
                if (project.name.match(new RegExp(value, 'i'))) {
                  projects.push({
                    action: {
                      name: 'loadProject',
                      data: {
                        name: project.name
                      }
                    },
                    text: project.name
                  });
                }
              }
              
              autocode.fuzzy.open({
                rows: projects,
                target: $('#popup input[name="name"]'),
                value: value
              });
              
              $('#popup input[name="name"]').attr('placeholder', 'Search GitHub...');
            };
            
            autocode.popup.open({
              title: 'Load Project',
              content: form.toString()
            });
          }
        });
      }
    });
    
    return;
  }
  
  autocode.unload.enable();
  
  $('.app, #init').fadeOut();
  
  $('#popup input[name="name"]').val(opts.name);
  
  autocode.fuzzy.close();
  $('#popup, #overlay').fadeOut(function() {
    autocode.popup.close();
  });
  
  autocode.repo = opts.name;
  
  var user = opts.name.split('/')[0];
  if (user == autocode.data.user.username) {
    $('#project .text').text(opts.name.split('/')[1]);
  } else {
    $('#project .text').text(opts.name);
  }
  
  if (opts.config) {
    autocode.project = opts.config;
    autocode.data.originalConfig = opts.config;
    
    autocode.data.current = {
      pin: true
    };
    autocode.data.engines = {};
    autocode.data.generators = {};
    autocode.data.schemas = {};
    autocode.imports = {};
    
    var requests = [];
    for (var import_name in autocode.project.imports) {
      requests.push(
        autocode.api.config.get({
          data: {
            repo: import_name
          },
          success: function(data) {
            var imported = this.url.split('?')[1];
            imported = autocode.query.search(imported);
            
            data.config = jsyaml.safeLoad(data.config)
            
            autocode.imports[imported.repo] = data.config;
            
            if (data.config.exports) {
              for (var export_name in data.config.exports) {
                switch (data.config.exports[export_name].type) {
                  case 'engine': {
                    autocode.data.engines[imported.repo.split('/')[1] + '.' + export_name] = JSON.parse(JSON.stringify(data.config.exports[export_name]));
                    break;
                  }
                  case 'generator': {
                    autocode.data.generators[imported.repo.split('/')[1] + '.' + export_name] = JSON.parse(JSON.stringify(data.config.exports[export_name]));
                    break;
                  }
                  case 'schema': {
                    autocode.data.schemas[imported.repo.split('/')[1] + '.' + export_name] = JSON.parse(JSON.stringify(data.config.exports[export_name]));
                    break;
                  }
                }
              }
            }
          }
        })
      );
    }
    
    $.when.apply(undefined, requests).done(function() {
      autocode.data.generators = autocode.object.sort(autocode.data.generators);
      autocode.data.engines = autocode.object.sort(autocode.data.engines);
      autocode.data.schemas = autocode.object.sort(autocode.data.schemas);
      
      $('#welcome').fadeOut(function() {
        $('#overview-tab').prop('href', 'overview');
        $('#imports-tab').prop('href', 'imports');
        $('#config-tab').prop('href', 'config');
        $('#output-tab').prop('href', 'output');
        $('#interfaces-tab').prop('href', 'interfaces');
        
        $('#overview-general-subtab').prop('href', 'overview/general');
        $('#overview-author-subtab').prop('href', 'overview/author');
        
        $('header').fadeIn();
        $('#menu-project').show();
        
        autocode.state['overview']();
        
        $('.app').fadeIn();
        
        if (opts.callback) {
          opts.callback();
        }
        
        autocode.resize.all();
      });
    });
    
    return;
  }
  
  autocode.api.config.get({
    data: {
      repo: opts.name
    },
    error: function(data) {
      $('#welcome').fadeOut(function() {
        $('#init').fadeIn();
      });
    },
    success: function(data) {
      var recent_projects = autocode.storage.get('recent', []);
      if (recent_projects.indexOf(opts.name) !== -1) {
        recent_projects.splice(
          recent_projects.indexOf(opts.name),
          1
        );
      }
      recent_projects.push(opts.name);
      if (recent_projects.length > 3) {
        recent_projects = recent_projects.splice(recent_projects.length - 3, 3);
      }
      autocode.storage.set('recent', recent_projects);
      
      autocode.action.updateRecent();
      
      var config = autocode.storage.get('config');
      if (config && config[autocode.repo]) {
        autocode.project = jsyaml.safeLoad(config[autocode.repo]);
      } else {
        autocode.project = jsyaml.safeLoad(data.config);
      }
      autocode.data.originalConfig = data.config;
      
      autocode.data.engines = {};
      autocode.data.generators = {};
      autocode.data.schemas = {};
      autocode.imports = {};
      
      var requests = [];
      for (var import_name in autocode.project.imports) {
        requests.push(
          autocode.api.config.get({
            data: {
              repo: import_name
            },
            success: function(data) {
              var imported = this.url.split('?')[1];
              imported = autocode.query.search(imported);
              
              var import_name = imported.repo.split('/');
              if (import_name.length > 1) {
                import_name = import_name[1];
              } else {
                import_name = import_name[0];
              }
              
              data.config = jsyaml.safeLoad(data.config)
              
              autocode.imports[import_name] = data.config;
              
              if (data.config.exports) {
                for (var export_name in data.config.exports) {
                  switch (data.config.exports[export_name].type) {
                    case 'engine': {
                      autocode.data.engines[imported.repo.split('/')[1] + '.' + export_name] = JSON.parse(JSON.stringify(data.config.exports[export_name]));
                      break;
                    }
                    case 'generator': {
                      autocode.data.generators[imported.repo.split('/')[1] + '.' + export_name] = JSON.parse(JSON.stringify(data.config.exports[export_name]));
                      break;
                    }
                    case 'schema': {
                      autocode.data.schemas[imported.repo.split('/')[1] + '.' + export_name] = JSON.parse(JSON.stringify(data.config.exports[export_name]));
                      break;
                    }
                  }
                }
              }
            }
          })
        );
      }
      
      $.when.apply(undefined, requests).done(function() {
        autocode.data.engines = autocode.object.sort(autocode.data.engines);
        autocode.data.generators = autocode.object.sort(autocode.data.generators);
        autocode.data.schemas = autocode.object.sort(autocode.data.schemas);
        
        $('#welcome').fadeOut(function() {
          document.title = autocode.repo + ' | Autocode';
          
          $('#overview-tab').prop('href', opts.name + '/overview');
          $('#imports-tab').prop('href', opts.name + '/imports');
          $('#exports-tab').prop('href', opts.name + '/exports');
          $('#outputs-tab').prop('href', opts.name + '/outputs');
          $('#interfaces-tab').prop('href', opts.name + '/interfaces');
          $('#scripts-tab').prop('href', opts.name + '/scripts');
          $('#config-tab').prop('href', opts.name + '/config');
          $('#output-tab').prop('href', opts.name + '/output');
          
          $('#overview-general-subtab').prop('href', opts.name + '/overview/general');
          $('#overview-author-subtab').prop('href', opts.name + '/overview/author');
                    
          $('header').fadeIn();
          $('#menu-project').show();
          
          autocode.state['overview']();
          
          $('.app').fadeIn();
          
          if (opts.callback) {
            opts.callback();
          }
          
          autocode.resize.all();
        });
      });
    }
  });
};;autocode.action.loadScript = function(opts) {
  opts = opts || {};
  
  autocode.data.current.script = opts.name || autocode.data.current.script;
  delete(autocode.data.current.commands);
  
  $('#scripts-content-container .table a').removeClass('selected');
  $('#scripts-' + autocode.data.current.script).addClass('selected');
  
  if (!autocode.project.scripts || !autocode.project.scripts[autocode.data.current.script]) {
    $('#scripts-commands').html('<h3>There are no commands for ' + autocode.data.current.script.toUpperCase() + '. <a onclick="autocode.action.addCommand()">Click here</a> to add one.</h3>');
    autocode.resize.all();
    return;
  }
  
  autocode.data.current.commands = autocode.project.scripts[autocode.data.current.script];
  
  $('#scripts-commands').replaceWith('<div id="scripts-commands"></div>');
  
  var commands = autocode.project.scripts[autocode.data.current.script];
  
  var command, command_html;
  for (var i in commands) {
    command = commands[i];
    
    command_html = '<div class="field move">';
    if (command.command) {
      command_html += '<code class="value" onclick="autocode.action.editCommand({ index: ' + i + ' })">';
      if (command.description) {
        command_html += '<span class="icon info-icon" data-hint="' + command.description + '"></span> ';
      }
      if (command.path) {
        command_html += '<span class="path">' + command.path + '$</span> ';
      }
      command_html += '<span class="code">' + command.command + '</div>';
      command_html += '</code>';
    } else {
      command_html += '<h3 class="value" onclick="autocode.action.editCommandTitle({ index: ' + i + ' })">';
      command_html += command.title;
      command_html += '</h3>';
    }
    command_html += '</div>';
    
    command_html = $(command_html);
    
    $('#scripts-commands').append(command_html);
  }
  
  dragula([$('#scripts-commands')[0]], {
    mirrorContainer: document.body
  }).on('drop', function() {
    autocode.project.scripts[autocode.data.current.script] = [];
    $('#scripts-commands .field').each(function() {
      var script = {};
      if ($(this).find('h3').length) {
        script.title = $(this).find('h3').text();
      } else {
        script.description = $(this).find('.info-icon').data('hint');
        script.command = $(this).find('code .code').text();
        if ($(this).find('.path').length) {
          script.path = $(this).find('.path').text().substr(0, $(this).find('.path').text().length-1);
        }
      }
      autocode.project.scripts[autocode.data.current.script].push(script);
    });
    autocode.action.loadScript();
  });
  
  autocode.hint.init();
  
  autocode.resize.all();
};;autocode.action.login = function() {
  var href ='https://github.com/login/oauth/authorize';
  href += '?client_id=' + autocode.data.auth.client_id;
  href += '&redirect_uri=' + autocode.data.auth.redirect_uri;
  href += '&scope=' + autocode.data.auth.scope;
  
  window.location.href = href;
};;autocode.action.logout = function() {
  autocode.loader.open();
  
  autocode.api.logout.get({
    success: function() {
      history.pushState(null, null, '/');
      window.location.reload();
    }
  });
};;autocode.action.newProject = function() {
  autocode.popup.open({
    title: 'Loading...'
  });
  autocode.popover.close();
  
  new formulator({
    formula: 'formulas/forms/NewProject.json',
    xhr: true,
    ready: function(form) {
      form.action = autocode.url.api('repos');
      autocode.popup.open({
        title: 'New Project',
        content: form.toString()
      });
    },
    success: function(data) {
      $('#popup, #overlay').fadeOut(function() {
        autocode.popup.close();
        
        autocode.action.loadProject({ name: data.name });
        
        $('#welcome').fadeOut(function() {
          $('.app').fadeIn();
        });
      });
    }
  });
};;autocode.action.outputsHide = function() {
  var checkbox = $('#outputs-content-container .checkbox');
  
  if (checkbox.hasClass('checked')) {
    $('#outputs-content-container .schema .field').each(function() {
      var input = $(this).find('input');
      if (input.val().length) {
        $(this).show();
      } else {
        $(this).hide();
      }
    });
  } else {
    $('#outputs-content-container .schema .field').show();
  }
};;autocode.action.outputsToggle = function() {
  var checkbox = $('#outputs-content-container .checkbox');
  
  if (checkbox.hasClass('checked')) {
    checkbox.removeClass('checked');
  } else {
    checkbox.addClass('checked');
  }
  
  autocode.action.outputsHide();
};;autocode.action.reboot = function() {
  // close popover
  autocode.popover.close();
  
  // update status
  autocode.status.offline();
  
  // hide usage
  $('#usage-on').hide();
  $('#usage-off').show();
  
  $('#target-icon .text').text('Rebooting...');
  
  autocode.api.reboot.post({
    error: function(data) {
      console.log(data);
    },
    success: function(data) {
      console.log(data);
    }
  });
};;autocode.action.reimage = function() {
  // close popover
  autocode.popover.close();
  
  // update status
  autocode.status.offline();
  
  // hide usage
  $('#usage-on').hide();
  $('#usage-off').show();
  
  // reset ip
  delete(autocode.ws.ip);
  
  $('#target-icon .text').text('Reimaging...');
  
  autocode.api.reimage.post({
    error: function(data) {
      console.log(data);
    },
    success: function(data) {
      console.log(data);
    }
  });
};;autocode.action.removeExport = function(opts) {
  opts = opts || {};
  
  if (!opts.confirm) {
    autocode.popup.open({
      title: 'Remove Export',
      content: '<div style="padding-bottom: 10px">Are you sure you want to remove the <b>' + autocode.data.current.export + '</b> export from this project?</div>'
        + '<button class="delete" onclick="autocode.action.removeExport({ confirm: true })">Yes, Remove Export</button> <button class="secondary" onclick="autocode.action.closePopup()">No, Keep Export</button>'
    });
    
    return;
  }
  
  autocode.popup.close();
  
  for (var generator_name in autocode.data.generators) {
    if (generator_name.split('.')[0] == autocode.data.current.export.split('/')[1]) {
      delete(autocode.data.generators[generator_name]);
    }
  }
  delete(autocode.project.exports[autocode.data.current.export]);
  delete(autocode.data.current.export);
  
  autocode.state['exports']();
};;autocode.action.removeImport = function(opts) {
  opts = opts || {};
  
  if (!opts.confirm) {
    autocode.popup.open({
      title: 'Remove Import',
      content: '<div style="padding-bottom: 10px">Are you sure you want to remove the <b>' + autocode.data.current.import + '</b> import from this project?</div>'
        + '<button class="delete" onclick="autocode.action.removeImport({ confirm: true })">Yes, Remove Import</button> <button class="secondary" onclick="autocode.action.closePopup()">No, Keep Import</button>'
    });
    
    return;
  }
  
  autocode.popup.close();
  
  for (var generator_name in autocode.data.generators) {
    if (generator_name.split('.')[0] == autocode.data.current.import.split('/')[1]) {
      delete(autocode.data.generators[generator_name]);
    }
  }
  delete(autocode.project.imports[autocode.data.current.import]);
  
  autocode.state['imports']();
};;autocode.action.removeInterface = function(opts) {
  opts = opts || {};
  
  if (!opts.confirm) {
    autocode.popup.open({
      title: 'Remove Interface',
      content: '<div style="padding-bottom: 10px">Are you sure you want to remove the <b>' + autocode.project.interfaces[autocode.data.current.interface].name + '</b> interface from this project?</div>'
        + '<button onclick="autocode.action.closePopup()">No, Keep Interface</button> <button class="secondary" onclick="autocode.action.removeInterface({ confirm: true })">Yes, Remove Interface</button>'
    });
    
    return;
  }
  
  autocode.popup.close();
  
  autocode.project.interfaces.splice(autocode.data.current.interface, 1);
  delete(autocode.data.current.interface);
  if (!autocode.project.interfaces.length) {
    delete(autocode.project.interfaces);
  }
  
  autocode.state['interfaces']();
};;autocode.action.removeOutput = function(opts) {
  opts = opts || {};
  
  if (!opts.confirm) {
    autocode.popup.open({
      title: 'Delete Output',
      content: 'Are you sure you want to delete this output? <button class="delete" onclick="autocode.action.removeOutput({ confirm: true })">Delete Output</button> <button class="secondary" onclick="autocode.action.closePopup()">Cancel</button>'
    });
    return;
  }
  
  autocode.project.outputs.splice(autocode.data.current.output, 1);

  delete(autocode.data.current.output);

  autocode.popup.close();

  autocode.state['outputs']({ output: autocode.object.firstKey(autocode.project.outputs) });
};;autocode.action.reset = function(opts) {
  opts = opts || {};
  
  if (!autocode.project) {
    return;
  }

  autocode.popover.close();
  
  if (opts.confirm) {
    // remove project from local storage
    var config = autocode.storage.get('config');
    if (!config || typeof(config) != 'object') {
      config = {};
    }
    delete(config[autocode.repo]);
    autocode.storage.set('config', config);
    
    autocode.project = jsyaml.safeLoad(autocode.data.originalConfig);
    if (autocode.data.current.tab == 'config') {
      $('#config-content .CodeMirror')[0].CodeMirror.setValue(autocode.data.originalConfig);
    }
    autocode.popup.close();
    
    return;
  }
  
  if (autocode.data.current.tab == 'config') {
    var value = $('#config-content .CodeMirror')[0].CodeMirror.getValue();
    autocode.project = jsyaml.safeLoad(value);
  }

  if (autocode.data.originalConfig == jsyaml.safeDump(autocode.project)) {
    autocode.popup.open({
      title: 'No Changes',
      content: 'There are no changes to your Autocode configuration.'
    });
    return;
  }

  var output;
  for (var output_i in autocode.project.outputs) {
    output = autocode.project.outputs[output_i];
    
    if (!autocode.data.generators[output.generator]) {
      autocode.popup.open({
        title: 'Validation Error',
        content: 'Generator does not exist for output: <b>' + output.generator + '</b>'
      });
      return;
    }
  }

  autocode.popup.open({
    title: 'Reset Project',
    content: '<div>Are you sure you want to discard all changes? <a href="#">Show Changes</a></div><div class="diff"></div><button class="delete" onclick="autocode.action.reset({ confirm: true })">Reset Project</button> <button class="secondary" onclick="autocode.action.closePopup()">Cancel</button>'
  });

  CodeMirror.MergeView($('#popup .diff')[0], {
    value: jsyaml.safeDump(autocode.project),
    orig: autocode.data.originalConfig,
    showDifferences: true,
    lineNumbers: true,
    mode: 'yaml',
    readOnly: true,
    revertButtons: false
  });
  
  autocode.resize.popup();
};;autocode.action.run = function() {
  if (autocode.data.current.running) {
    autocode.data.current.running = false;
    
    autocode.ws.io.emit('stop', {
      config: autocode.project,
      project: autocode.repo.split('/')[1],
      user: autocode.repo.split('/')[0]
    });
    
    return;
  }
  
  if (!autocode.project.outputs || !autocode.project.imports) {
    if (!autocode.project.imports) {
      autocode.popup.open({
        title: 'Unable to Build Project',
        content: '<div style="padding-bottom: 10px">You haven\'t imported any modules yet.</div>'
          + '<a class="button" href="imports">Import a Module</a> <button class="secondary" onclick="autocode.action.closePopup()">Cancel</button>'
      });
    } else {
      autocode.popup.open({
        title: 'Unable to Build Project',
        content: '<div style="padding-bottom: 10px">You haven\'t added any outputs to your Autocode configuration.</div>'
          + '<a class="button" href="config">Add an Output</a> <button class="secondary" onclick="autocode.action.closePopup()">Cancel</button>'
      });
    }
    
    return;
  }
  
  autocode.data.current.running = true;
    
  if (autocode.data.current.tab == 'config') {
    autocode.project = jsyaml.safeLoad($('#config-content .CodeMirror')[0].CodeMirror.getValue());
  }
  
  autocode.ws.io.emit('run', {
    config: autocode.project,
    project: autocode.repo.split('/')[1],
    user: autocode.repo.split('/')[0]
  });
};;autocode.action.saveProject = function() {
  if (!autocode.repo) {
    return;
  }
  
  var config = autocode.storage.get('config');
  if (!config || typeof(config) != 'object') {
    config = {};
  }
  
  switch (autocode.data.current.tab) {
    case 'config': {
      autocode.project = jsyaml.safeLoad($('#config-content .CodeMirror')[0].CodeMirror.getValue());
      break;
    }
    case 'exports': {
      var Export = autocode.project.exports[autocode.data.current.export];
      switch (Export.type) {
        case 'engine': {
          autocode.project.exports[autocode.data.current.export].engine = $('#exports-content .CodeMirror')[0].CodeMirror.getValue();
          break;
        }
        case 'helper': {
          autocode.project.exports[autocode.data.current.export].helper = $('#exports-content .CodeMirror')[0].CodeMirror.getValue();
          break;
        }
        case 'processor': {
          autocode.project.exports[autocode.data.current.export].processor = $('#exports-content .CodeMirror')[0].CodeMirror.getValue();
          break;
        }
        case 'generator': {
          autocode.project.exports[autocode.data.current.export].template = $('#exports-content .CodeMirror')[0].CodeMirror.getValue();
          break;
        }
        case 'schema': {
          autocode.project.exports[autocode.data.current.export].schema = jsyaml.safeLoad($('#exports-content .CodeMirror')[0].CodeMirror.getValue());
          break;
        }
        case 'spec': {
          autocode.project.exports[autocode.data.current.export].spec = jsyaml.safeLoad($('#exports-content .CodeMirror')[0].CodeMirror.getValue());
          break;
        }
        case 'transformer': {
          autocode.project.exports[autocode.data.current.export].transformer = $('#exports-content .CodeMirror')[0].CodeMirror.getValue();
          break;
        }
      }
      break;
    }
    case 'outputs': {
      autocode.project.outputs[autocode.data.current.output].spec = jsyaml.safeLoad($('#outputs-content .CodeMirror')[0].CodeMirror.getValue());
      break;
    }
  }
  
  config[autocode.repo] = jsyaml.safeDump(autocode.project);
  
  autocode.storage.set('config', config);
};;autocode.action.setEditorTheme = function(opts) {
  var theme = opts.name.toLowerCase();
  
  $('#popup input[name="theme"]').val(theme);
  
  if (!$('#' + theme + '-theme').length) {
    $('head').append('<link id="' + theme + '-theme" href="components/CodeMirror/theme/' + theme + '.css" rel="stylesheet" type="text/css" />');
  }
  $('#popup .CodeMirror')[0].CodeMirror.setOption('theme', theme);

  autocode.fuzzy.close();
  
  autocode.resize.popup();
};;autocode.action.setExportFormat = function(opts) {
  $('#popup input[name="format"]').val(opts.name);
  autocode.fuzzy.close();
};;autocode.action.settings = function() {
  autocode.popover.close();
  
  autocode.loader.open({ title: 'Loading Settings...' });
  
  new formulator({
    formula: 'formulas/forms/Settings.json',
    xhr: true,
    ready: function(form) {
      form.action = autocode.url.api('repos');
      form.fields.email.placeholder = autocode.data.user.email;
      form.fields.name.placeholder = autocode.data.user.name;
      form.fields.url.placeholder = autocode.data.user.website;
      form.fields.copyright.placeholder = new Date().getFullYear();
      if (autocode.data.user.company) {
        form.fields.copyright.placeholder += ' ' + autocode.data.user.company;
      } else if (autocode.data.user.name) {
        form.fields.copyright.placeholder += ' ' + autocode.data.user.name;
      }
      
      autocode.popup.open({
        title: 'Settings',
        content: '<div style="padding-bottom: 10px">Note: Defaults to GitHub info.</div>' + form.toString()
      });
    },
    success: function(data) {
      $('#popup, #overlay').fadeOut(function() {
        autocode.popup.close();
        
        $('#welcome').fadeOut(function() {
          $('.app').fadeIn();
        });
      });
    }
  });
};;autocode.action.share = function() {
  if (autocode.data.current.tab == 'config') {
    autocode.project = jsyaml.safeLoad($('#config-content .CodeMirror')[0].CodeMirror.getValue());
  }
  
  var value = jsyaml.safeDump(autocode.project);
  value = btoa(value);
  
  autocode.popup.open({
    title: 'Share Configuration',
    content: '<input id="popup-copy-target" type="text" value="' + location.protocol + '//' + location.host + '/?config=' + value + '" /><button class="copy-button" data-clipboard-target="popup-copy-target">Copy</button> <button class="secondary" onclick="autocode.action.closePopup()">Close</button>'
  });
  
  var client = new ZeroClipboard($('.copy-button')[0]);
};;autocode.action.stop = function() {
  if (!autocode.project.outputs || !autocode.project.imports) {
    if (!autocode.project.imports) {
      autocode.popup.open({
        title: 'Unable to Stop Project',
        content: '<div style="padding-bottom: 10px">You haven\'t imported any modules yet.</div>'
          + '<a class="button" href="imports">Import a Module</a> <button class="secondary" onclick="autocode.action.closePopup()">Cancel</button>'
      });
    } else {
      autocode.popup.open({
        title: 'Unable to Stop Project',
        content: '<div style="padding-bottom: 10px">You haven\'t added any outputs to your Autocode configuration.</div>'
          + '<a class="button" href="config">Add an Output</a> <button class="secondary" onclick="autocode.action.closePopup()">Cancel</button>'
      });
    }
    
    return;
  }
  
  if (autocode.data.current.tab == 'config') {
    autocode.project = jsyaml.safeLoad($('#config-content .CodeMirror')[0].CodeMirror.getValue());
  }
  
  autocode.ws.io.emit('stop', {
    config: autocode.project,
    project: autocode.repo.split('/')[1],
    user: autocode.repo.split('/')[0]
  });
};;autocode.action.toggleColumn = function(id, column, opts) {
  opts = opts || {};
  autocode.data.current.column = column;
  if ($(window).width() > autocode.mobile.minWidth) {
    $('#content').css({
      marginLeft: ''
    });
  } else if (opts.animated === false) {
    $('#content').css({
      marginLeft: $(window).width() * -(column - 1)
    });
  } else {
    $('#content').animate({
      marginLeft: $(window).width() * -(column - 1)
    });
  }
};;autocode.action.toggleConsole = function() {
  // close popover
  autocode.popover.close();
  
  // toggle console
  var showConsole = autocode.storage.get('showConsole');
  if (showConsole) {
    autocode.storage.remove('showConsole');
    $('#console').hide();
  } else {
    autocode.storage.set('showConsole', true);
    $('#console').show();
  }
  
  autocode.resize.all(['config','content','footer']);
  
  $('#console .content').scrollTop($('#console .content')[0].scrollHeight);
};;autocode.action.toggleMenu = function() {
  var rows = [];
  if (autocode.data.user) {
    if (autocode.project) {
      rows.push({
        action: 'saveProject',
        text: 'Save Project (Ctrl+Shift+S)',
        icon: 'save-icon'
      });
      rows.push({
        text: 'Close Project (Ctrl+Shift+W)',
        icon: 'close-icon',
        state: '/'
      });
      rows.push({
        action: 'commitProject',
        text: 'Commit Project (Ctrl+Shift+M)',
        style: 'divider',
        icon: 'commit-icon'
      });
      rows.push({
        text: 'Reset Project (Ctrl+Shift+T)',
        icon: 'close-icon',
        state: '/'
      });
      rows.push({
        action: 'githubRepo',
        icon: 'login-icon',
        text: 'View on GitHub'
      });
    }
    
    rows.push({
      action: 'newProject',
      text: 'New Project (Ctrl+Shift+N)',
      icon: 'add-icon',
      style: 'divider'
    });
    rows.push({
      action: 'loadProject',
      text: 'Open Project (Ctrl+Shift+O)',
      icon: 'load-icon'
    });
  } else {
    rows.push({
      action: 'login',
      text: 'Login with GitHub',
      icon: 'login-icon'
    });
  }
  rows.push({
    text: 'Take a Tour',
    icon: 'tour-icon',
    state: 'tour',
    style: 'divider'
  });
  
  autocode.popover.toggle({
    rows: rows,
    left: $('#project').offset().left,
    style: 'table',
    target: $('#project'),
    top: $('#main').outerHeight()
  });
  
  autocode.initState();
};;autocode.action.toggleMore = function() {
  autocode.popover.toggle({
    rows: [
      {
        action: 'halt',
        icon: 'halt-icon',
        text: 'Halt Process'
      },
      {
        action: 'clean',
        icon: 'clean-icon',
        style: 'divider',
        text: 'Clean Project'
      },
      {
        action: 'upgrade',
        icon: 'upgrade-icon',
        text: 'Upgrade Project'
      },
      {
        action: 'toggleConsole',
        icon: 'console-icon',
        style: 'divider',
        text: $('#console').is(':visible')
          ? 'Hide Console'
          : 'Show Console'
      }
    ],
    left: $('#more-icon').offset().left,
    style: 'table',
    target: $('#more-icon'),
    top: $('#main').outerHeight()
  });
  
  autocode.initState();
};;autocode.action.togglePin = function() {
  if (autocode.data.current.pin) {
    autocode.data.current.pin = false;
    $('#pin-icon').removeClass('active');
  } else {
    autocode.data.current.pin = true;
    $('#pin-icon').addClass('active');
    $('#console .content').scrollTop($('#console .content')[0].scrollHeight);
  }
};;autocode.action.toggleSection = function(name, section) {
  if (autocode.data.current.tab == 'config') {
    var value = $('#config-content .CodeMirror')[0].CodeMirror.getValue();
    autocode.project = jsyaml.safeLoad(value);
  }
  
  autocode.data.current.tab = name;
  autocode.data.current.subtab = section;
  
  $('.content, .tab').removeClass('selected');
  $('#' + name + '-content, #' + name + '-tab').addClass('selected');
  if (section) {
    $('.subcontent').removeClass('selected');
    $('#' + name + '-' + section + '-content').addClass('selected');
    $('#' + name + '-content .subtab').removeClass('selected');
    $('#' + name + '-' + section + '-subtab').addClass('selected');
  }
};;autocode.action.toggleTarget = function() {
  if (autocode.data.current.tab == 'config') {
    var config = $('#config-content .CodeMirror')[0].CodeMirror.getValue();
    autocode.project = jsyaml.safeLoad(config);
  }
  
  var rows = [];
  
  if (autocode.project.interfaces && autocode.project.interfaces.length) {
    var state, port, path, params, name;
    for (var i = 0; i < autocode.project.interfaces.length; i++) {
      state = 'http://' + autocode.ws.ip;
      port = autocode.project.interfaces[i].port;
      path = autocode.project.interfaces[i].path;
      params = autocode.project.interfaces[i].params;
      name = autocode.project.interfaces[i].name;
      
      if (port) {
        if (typeof(port) == 'string' && port.substr(0, 1) == '$' && autocode.project[port.substr(1)]) {
          port = autocode.project[port.substr(1)];
        }
        state += ':' + port;
      }
      if (path) {
        if (typeof(path) == 'string' && path.substr(0, 1) == '$' && autocode.project[path.substr(1)]) {
          path = autocode.project[path.substr(1)];
        }
        state += path;
      }
      if (params) {
        if (typeof(params) == 'string' && params.substr(0, 1) == '$' && autocode.project[params.substr(1)]) {
          params = autocode.project[params.substr(1)];
        }
        state += '?' + autocode.query.build(params);
      }
      if (typeof(name) == 'string' && name.substr(0, 1) == '$' && autocode.project[name.substr(1)]) {
        name = autocode.project[name.substr(1)];
      }
      
      rows.push({
        icon: 'interface-icon',
        state: state,
        text: name
      });
    }
  }
  
  rows.push({
    action: 'reboot',
    icon: 'reboot-icon',
    style: 'divider',
    text: 'Reboot'
  });
  rows.push({
    action: 'reimage',
    icon: 'reimage-icon',
    text: 'Reimage'
  });
  
  autocode.popover.toggle({
    rows: rows,
    left: $('#target-icon').offset().left,
    style: 'table',
    target: $('#user'),
    top: $('#main').outerHeight()
  });
  
  autocode.initState();
};;autocode.action.toggleUser = function() {
  if (autocode.data.user) {
    autocode.popover.toggle({
      rows: [
        /*
        {
          icon: 'settings-icon',
          state: 'user/settings',
          text: 'Settings'
        },
        */
        {
          action: 'githubUser',
          icon: 'login-icon',
          text: 'View on GitHub'
        },
        {
          action: 'editorSettings',
          icon: 'editor-icon',
          style: 'divider',
          text: 'Editor Settings'
        },
        {
          action: 'logout',
          icon: 'logout-icon',
          style: 'divider',
          text: 'Logout'
        }
      ],
      left: 0,
      style: 'table',
      target: $('#user'),
      top: $('#main').outerHeight()
    });
  } else {
    autocode.action.login();
  }
  
  autocode.initState();
};;autocode.action.update = function() {
  if (!autocode.project.outputs || !autocode.project.imports) {
    if (!autocode.project.imports) {
      autocode.popup.open({
        title: 'Unable to Update Project',
        content: '<div style="padding-bottom: 10px">You haven\'t imported any modules yet.</div>'
          + '<a class="button" href="imports">Import a Module</a> <button class="secondary" onclick="autocode.action.closePopup()">Cancel</button>'
      });
    } else {
      autocode.popup.open({
        title: 'Unable to Update Project',
        content: '<div style="padding-bottom: 10px">You haven\'t added any outputs to your Autocode configuration.</div>'
          + '<a class="button" href="config">Add an Output</a> <button class="secondary" onclick="autocode.action.closePopup()">Cancel</button>'
      });
    }
    
    return;
  }
  
  if (autocode.data.current.tab == 'config') {
    autocode.project = jsyaml.safeLoad($('#config-content .CodeMirror')[0].CodeMirror.getValue());
  }
  
  autocode.ws.io.emit('update', {
    config: autocode.project,
    project: autocode.repo.split('/')[1],
    user: autocode.repo.split('/')[0]
  });
};;autocode.action.updateRecent = function() {
  if (!autocode.data.user) {
    return;
  }
  
  var recent_projects = autocode.storage.get('recent', []).reverse();
  if (recent_projects.length) {
    var html = '';
    for (var i = 0; i < recent_projects.length; i++) {
      html += '<a href="' + recent_projects[i] + '">' + recent_projects[i] + '</a>';
    }
    $('#welcome .recent .projects').html(html);
    
    $('#welcome .recent').show();
    
    autocode.initState();
    
  } else {
    $('#welcome .recent').hide();
  }
};;autocode.action.upgrade = function() {
  // close popover
  autocode.popover.close();
  
  autocode.ws.io.emit('upgrade');
};;;autocode.api.auth = {
  uri: 'auth',
  init: function(opts) {
    opts.uri = autocode.api.auth.uri;
    return opts;
  },
  get: function(opts) {
    opts = autocode.api.auth.init(opts);
    return autocode.api.get(opts);
  }
};;autocode.api.box = {
  uri: 'box',
  init: function(opts) {
    opts.uri = autocode.api.box.uri;
    return opts;
  },
  post: function(opts) {
    opts = autocode.api.box.init(opts);
    return autocode.api.post(opts);
  }
};;autocode.api.build = {
  url: 'http://localhost:12345',
  uri: 'build',
  init: function(opts) {
    opts.uri = autocode.api.build.uri;
    return opts;
  },
  post: function(opts) {
    opts = autocode.api.build.init(opts);
    return autocode.api.post(opts);
  }
};;autocode.api.config = {
  uri: 'config?provider=1',
  init: function(opts) {
    opts.uri = autocode.api.config.uri;
    return opts;
  },
  get: function(opts) {
    opts = autocode.api.config.init(opts);
    return autocode.api.get(opts);
  },
  post: function(opts) {
    opts = autocode.api.config.init(opts);
    return autocode.api.post(opts);
  }
};;autocode.api.generate = {
  uri: 'generate',
  init: function(opts) {
    opts.uri = autocode.api.generate.uri;
    return opts;
  },
  post: function(opts) {
    opts = autocode.api.generate.init(opts);
    return autocode.api.post(opts);
  }
};;autocode.api.init = {
  uri: 'init',
  init: function(opts) {
    opts.uri = autocode.api.init.uri;
    return opts;
  },
  post: function(opts) {
    opts = autocode.api.init.init(opts);
    return autocode.api.post(opts);
  }
};;autocode.api.login = {
  uri: 'login',
  init: function(opts) {
    opts.uri = autocode.api.login.uri;
    return opts;
  },
  post: function(opts) {
    opts = autocode.api.login.init(opts);
    return autocode.api.post(opts);
  }
};;autocode.api.logout = {
  uri: 'logout',
  init: function(opts) {
    opts.uri = autocode.api.logout.uri;
    return opts;
  },
  get: function(opts) {
    opts = autocode.api.logout.init(opts);
    return autocode.api.get(opts);
  }
};;autocode.api.modules = {
  uri: 'modules',
  init: function(opts) {
    opts.uri = autocode.api.modules.uri;
    return opts;
  },
  get: function(opts) {
    opts = autocode.api.modules.init(opts);
    return autocode.api.get(opts);
  }
};;autocode.api.readme = {
  uri: 'readme?provider=1',
  init: function(opts) {
    opts.uri = autocode.api.readme.uri;
    return opts;
  },
  get: function(opts) {
    opts = autocode.api.readme.init(opts);
    return autocode.api.get(opts);
  }
};;autocode.api.reboot = {
  uri: 'reboot',
  init: function(opts) {
    opts.uri = autocode.api.reboot.uri;
    return opts;
  },
  post: function(opts) {
    opts = autocode.api.reboot.init(opts);
    return autocode.api.post(opts);
  }
};;autocode.api.reimage = {
  uri: 'reimage',
  init: function(opts) {
    opts.uri = autocode.api.reimage.uri;
    return opts;
  },
  post: function(opts) {
    opts = autocode.api.reimage.init(opts);
    return autocode.api.post(opts);
  }
};;autocode.api.releases = {
  uri: 'releases',
  init: function(opts) {
    opts.uri = autocode.api.releases.uri;
    return opts;
  },
  get: function(opts) {
    opts = autocode.api.releases.init(opts);
    return autocode.api.get(opts);
  }
};;autocode.api.repos = {
  uri: 'repos',
  init: function(opts) {
    opts.uri = autocode.api.repos.uri;
    return opts;
  },
  get: function(opts) {
    opts = autocode.api.repos.init(opts);
    return autocode.api.get(opts);
  }
};;autocode.api.run = {
  uri: 'run',
  init: function(opts) {
    opts.uri = autocode.api.run.uri;
    return opts;
  },
  post: function(opts) {
    opts = autocode.api.run.init(opts);
    return autocode.api.post(opts);
  }
};;autocode.api.user = {
  uri: 'user',
  init: function(opts) {
    opts.uri = autocode.api.user.uri;
    return opts;
  },
  get: function(opts) {
    opts = autocode.api.user.init(opts);
    return autocode.api.get(opts);
  }
};;;autocode.element.input = {
  html: function(opts) {
    var input = $(document.createElement('input'));
    
    if (opts.autocomplete === true) {
      input.attr('autocomplete', 'on');
    } else if (opts.autocomplete === false) {
      input.attr('autocomplete', 'off');
    }
    
    if (opts.class) {
      input.addClass(opts.class);
    }
    
    if (opts.css) {
      input.css(opts.css);
    }
    
    if (opts.event) {
      var e;
      for (var event_name in opts.event) {
        e = opts.event[event_name];
        input.bind(event_name, e);
      }
    }
    
    if (opts.name) {
      input.attr('name', opts.name);
    }
    
    if (opts.placeholder) {
      input.attr('placeholder', opts.placeholder);
    }
    
    if (opts.value) {
      input.val(opts.value);
    }
    
    return input;
  }
};;autocode.element.radio = {
  event: {
    click: function() {
      var field = $(this);
      var selected = $(this).hasClass('selected');
      
      field.parent().find('button').removeClass('selected');
      
      if (!selected) {
        field.addClass('selected');
        field.parent().find('input[type="hidden"]').val(field.val());
      } else {
        field.parent().find('input[type="hidden"]').val('');
      }
    }
  },
  html: function(opts) {
    var span = $(document.createElement('span'));
    span.addClass('radio');
    
    var button, option_text;
    for (var option_value in opts.options) {
      option_text = opts.options[option_value];
      
      button = $(document.createElement('button'));
      if ((opts.value || opts.value === false) && opts.value.toString() == option_value.toString()) {
        button.addClass('selected');
      }
      if ((opts.defaultValue || opts.defaultValue === false) && opts.defaultValue.toString() == option_value.toString()) {
        button.addClass('default');
      }
      button.attr('type', 'button');
      button.click(autocode.element.radio.event.click);
      button.text(option_text);
      button.val(option_value);
      span.append(button);
    }
    
    var input = $(document.createElement('input'));
    input.attr('name', opts.name);
    input.attr('type', 'hidden');
    input.val(opts.value || opts.value === false ? opts.value : '');
    span.append(input);
    
    return span;
  }
};;;autocode.resize.body = function() {
  var body = $('body');
  body.css({
    marginTop: $('header').outerHeight()
  });
};;autocode.resize.config = function() {
  var config = $('#config-content .CodeMirror');
  config.css({
    height: $(window).height() - $('header').outerHeight() - $('footer').outerHeight() - $('#console').visibleHeight()
  });
};;autocode.resize.content = function() {
  if ($(window).width() > autocode.mobile.minWidth) {
    $('#content').css({
      height: $(window).height() - $('header').outerHeight() - $('footer').outerHeight() - $('#console').visibleHeight(),
      width: $(window).width()
    });
  } else {
    $('#content').css({
      height: $(window).height() - $('header').outerHeight() - $('footer').outerHeight() - $('#console').visibleHeight(),
      width: $(window).width() * ($('.column:visible').length ? $('.column:visible').length : 1)
    });
  }
  if ($(window).outerWidth() < autocode.mobile.minWidth) {
    $('#content .column').css({
      width: $(window).width()
    });
    $('.content-center').css({
      width: $(window).width()
    });
    $('#content .column').each(function() {
      $(this).css({
        left: $(window).width() * $(this).index()
      })
    });
  } else {
    $('#content .column').css({
      width: ''
    });
    $('.content-center').each(function() {
      $(this).css({
        width: $(window).width() - $(this).prev().visibleWidth() - $(this).next().visibleWidth()
      });
    });
    $('#content .column').each(function() {
      $(this).css({
        left: ''
      })
    });
  }
  if ($(window).width() > autocode.mobile.minWidth) {
    $('.content-left, .content-center, .content-right').css({
      height: $(window).height() - $('header').outerHeight() -  $('footer').outerHeight() - $('#console').visibleHeight()
    });
  } else {
    $('.content-left, .content-center, .content-right').css({
      height: ''
    });
  }
};;autocode.resize.exports = function() {
  var exports = $('#exports-content .CodeMirror');
  exports.css({
    height: $(window).height() - $('header').outerHeight() - $('footer').outerHeight() - $('#exports-tabs').outerHeight() - 20
  });
};;autocode.resize.exportsInit = function() {
  var exports_init = $('#exports-init');
  exports_init.css({
    left: ($(window).width() - exports_init.outerWidth()) / 2,
    top: ($(window).height() - exports_init.outerHeight()) / 2
  });
};;autocode.resize.footer = function() {
  var footer = $('footer');
  footer.css('bottom', $('#console').visibleHeight());
};;autocode.resize.fuzzy = function() {
  var fuzzy = $('#fuzzy');
  var fuzzy_target = fuzzy.data('target');
  if (fuzzy.length && fuzzy_target.length) {
    var fuzzy_box_shadow = '0px 5px 10px #999';
    var fuzzy_max_height = $(window).height() - (fuzzy_target.outerHeight() + fuzzy_target.offset().top) - 50;
    var fuzzy_max_top = fuzzy_target.outerHeight() + fuzzy_target.offset().top;
    
    if (fuzzy_max_height > 300) {
      fuzzy_max_height = 300;
      fuzzy.css('maxHeight', fuzzy_max_height);
    } else if (fuzzy_max_height < 100) {
      fuzzy_box_shadow = '0px -5px 10px #999';
      fuzzy_max_height = 300;
      fuzzy.css('maxHeight', fuzzy_max_height);
      fuzzy_max_top = fuzzy_target.offset().top - fuzzy.outerHeight();
    }
    
    fuzzy.css({
      boxShadow: fuzzy_box_shadow,
      left: fuzzy_target.offset().left,
      top: fuzzy_max_top,
      width: fuzzy_target.outerWidth()
    });
  }
};;autocode.resize.header = function() {
  $('header').css('width', $(window).width());
};;autocode.resize.hint = function() {
  var hint = $('#hint');
  var hint_arrow = $('#hint-arrow');
  var hint_target = hint.data('target');
  if (typeof(hint_target) == 'function') {
    hint_target = hint_target();
  }
  
  if (!hint.length || !hint_target.length) {
    return;
  }
  
  var hint_minTop = hint.data('minTop') !== undefined ? hint.data('minTop') : ($('header').outerHeight() + 10);
  var hint_offsetTop = hint.data('offsetTop') || 0;
  var hint_text = hint.data('originalText');
  var hint_width = hint.outerWidth();

  var hint_left = hint_target.offset().left - ((hint.outerWidth() - hint_target.outerWidth()) / 2);
  if (hint_left < 10) {
    hint_left = 10;
  } else if (hint_left + hint_width > $(window).width()) {
    hint_left = $(window).width() - hint_width - 10;
  }
  
  var hint_top;
  if (hint.data('top')) {
    hint_top = hint.data('top');
  } else {
    hint_top = hint_target.offset().top + hint_target.outerHeight() + hint_offsetTop + $(window).scrollTop() + 10;
    if (hint_top < hint_minTop) {
      hint_top = hint_minTop;
      if (hint.data('scrollUp')) {
        hint_text = hint.data('scrollUp');
      }
    }
  }
  
  hint.css({
    left: hint_left,
    top: hint_top
  });
  hint.text(hint_text);
  
  hint_arrow.css({
    left: hint_target.offset().left + ((hint_target.outerWidth() - hint_arrow.outerWidth()) / 2),
    top: hint_top - 14
  });
};;autocode.resize.init = function() {
  var init = $('#init');
  init.css({
    left: ($(window).width() - init.outerWidth()) / 2,
    top: ($(window).height() - init.outerHeight()) / 2
  });
};;autocode.resize.interfacesInit = function() {
  var interfaces_init = $('#interfaces-init');
  interfaces_init.css({
    left: ($(window).width() - interfaces_init.outerWidth()) / 2,
    top: ($(window).height() - interfaces_init.outerHeight()) / 2
  });
};;autocode.resize.loader = function() {
  var loader = $('#loader');
  loader.css({
    left: ($(window).width() - loader.outerWidth()) / 2,
    top: (($(window).height() - $('#console').visibleHeight()) - loader.outerHeight()) / 2
  });
};;autocode.resize.output = function() {
  var output = $('#output-content .CodeMirror');
  output.css({
    height: $(window).height()
      - $('header').outerHeight()
      - $('footer').outerHeight()
      - $('#console').visibleHeight()
  });
};;autocode.resize.outputInit = function() {
  var output_init = $('#output-init');
  output_init.css({
    left: ($(window).width() - output_init.outerWidth()) / 2,
    top: ($(window).height() - output_init.outerHeight()) / 2
  });
};;autocode.resize.outputs = function() {
  var outputs = $('#outputs-content .CodeMirror');
  outputs.css({
    height: $(window).height() - $('header').outerHeight() - $('footer').outerHeight() - $('#outputs-tabs').outerHeight() - 20
  });
};;autocode.resize.outputsInit = function() {
  var outputs_init = $('#outputs-init');
  outputs_init.css({
    left: ($(window).width() - outputs_init.outerWidth()) / 2,
    top: ($(window).height() - outputs_init.outerHeight()) / 2
  });
};;autocode.resize.overlay = function() {
  var overlay = $('#overlay');
  overlay.css({
    height: $(window).height(),
    width: $(window).width()
  });
};;autocode.resize.popup = function() {
  var popup = $('#popup');
  $('#popup .content').css('maxHeight', $(window).height() - 100);
  popup.css({
    left: ($(window).width() - popup.outerWidth()) / 2,
    top: ($(window).height() - popup.outerHeight()) / 2
  });
};;autocode.resize.welcome = function() {
  var welcome = $('#welcome');
  if (welcome.outerHeight() > $(window).height() && welcome.outerWidth() > autocode.mobile.minWidth) {
    welcome.css({
      left: ($(window).width() - welcome.outerWidth()) / 2,
      position: 'absolute',
      top: $('header').outerHeight() + 20
    });
  } else {
    welcome.css({
      left: ($(window).width() - welcome.outerWidth()) / 2,
      position: '',
      top: (($(window).height() - $('footer').visibleHeight() - $('#console').visibleHeight()) - welcome.outerHeight()) / 2
    });
  }
};;;autocode.state['config'] = function() {
  autocode.popup.close();
  
  autocode.action.toggleColumn('config-content', 1, { animated: false });
  autocode.action.toggleSection('config');
  
  var code_mirror = $('#config-content .CodeMirror');
  var value = jsyaml.safeDump(autocode.project);
    
  if (!code_mirror.length) {
    var editor = CodeMirror.fromTextArea($('#config-content textarea')[0], {
      lineNumbers: autocode.editor.lineNumbersEnabled(),
      mode: 'yaml',
      theme: autocode.editor.getTheme()
    });
    
    code_mirror = $('#config-content .CodeMirror')
    code_mirror[0].CodeMirror.setValue(value);
    
    $('.CodeMirror-scroll').scrollTop(2);
    
  } else {
    code_mirror[0].CodeMirror.setValue(value);
  }
  
  setTimeout(function() { $('#config-content .CodeMirror')[0].CodeMirror.refresh() }, 0);
  
  autocode.resize.all();
};;;autocode.state['exports'] = function() {
  if (autocode.project.exports) {
    $('#exports-content-container .table a').slice(1).remove();
    
    var exported;
    for (var export_name in autocode.project.exports) {
      exported = autocode.project.exports[export_name];
      
      $('#exports-content-container .table').append(
        '<a class="file" id="exports-' + export_name + '" onclick="autocode.action.loadExport({ name: \'' + export_name + '\' })">'
          + '<span class="image">'
            + '<span class="icon ' + exported.type.toLowerCase() + '-icon"></span>'
          + '</span>'
          + '<span class="info">'
            + '<span class="name">' + export_name + '</span>'
            + '<span class="generator">' + exported.type + '</span>'
          + '</span>'
        + '</a>'
      );
    }
    
    autocode.initState()
    
    $('#exports-init').hide();
    $('#exports-content-container').show();
    
    autocode.action.loadExport();
    
  } else {
    $('#exports-init').show();
    $('#exports-content-container').hide();
  }
  
  autocode.action.toggleSection('exports');
};;;autocode.state['exports/export'] = function(opts) {
  var exported = autocode.project.exports[opts.export];
  
  $('#exports-content input[name="name"]').val(opts.export);
  $('#exports-content input[name="description"]').val(exported.description);
  $('#exports-content input[name="type"]').val(exported.type);
  
  var mode = 'yaml';
  var value;
  if (exported.type == 'generator') {
    mode = exported.format || 'text';
    
    $('#exports-content input[name="engine"]').val(exported.engine);
    $('#exports-content input[name="filename"]').val(exported.filename);
    $('#exports-content input[name="schema"]').val(exported.schema);
    $('#exports-content input[name="format"]').val(mode);
    
    $('#exports-content .engine-field, #exports-content .filename-field, #exports-content .format-field, #exports-content .schema-field').show();
    
    value = exported.template;
    
  } else {
    $('#exports-content input[name="engine"]').val('');
    $('#exports-content input[name="filename"]').val('');
    $('#exports-content input[name="schema"]').val('');
    
    $('#exports-content .engine-field, #exports-content .filename-field, #exports-content .format-field, #exports-content .schema-field').hide();
    
    value = jsyaml.safeDump(exported.schema);
  }
  
  var code_mirror = $('#exports-content .CodeMirror');
  if (!code_mirror.length) {
    var editor = CodeMirror.fromTextArea($('#exports-content textarea')[0], {
      lineNumbers: true,
      mode: mode
    });
    
    code_mirror = $('#exports-content .CodeMirror');
    code_mirror[0].CodeMirror.setValue(value);
    
    $('.CodeMirror-scroll').scrollTop(2);
    
  } else {
    code_mirror[0].CodeMirror.setOption('mode', mode);
    code_mirror[0].CodeMirror.setValue(value);
  }
  
  $('#exports-content form .button').attr('href', 'exports/export/save?export=' + opts.export);
};;autocode.state['exports/export/save'] = function(opts) {
  if ($('#exports-content input[name="name"]').val() != opts.export) {
    var export_name = $('#exports-content input[name="name"]').val();
    autocode.project.exports[export_name] = JSON.parse(JSON.stringify(autocode.project.exports[opts.export]));
    delete(autocode.project.exports[opts.export]);
  } else {
    var export_name = opts.export;
  }
  
  //autocode.project.exports[opts.export].filename = $('#exports-content input[name="filename"]').val();
  autocode.project.exports[export_name].description = $('#exports-content input[name="description"]').val();
  autocode.project.exports[export_name].type = $('#exports-content input[name="type"]').val();
  if (autocode.project.exports[export_name].type == 'generator' && $('#exports-content input[name="filename"]').val().length) {
    autocode.project.exports[export_name].filename = $('#exports-content input[name="filename"]').val();
  } else {
    delete(autocode.project.exports[export_name].filename);
  }
  
  autocode.state['project/save']();
};;autocode.state['/'] = function() {
  autocode.action.closeProject();
};;autocode.state['imports'] = function(opts) {
  opts = opts || {};
  
  autocode.popup.close();
  
  autocode.action.toggleColumn('imports-content', 1, {
    animated: autocode.data.current.tab == 'imports'
      ? true
      : false
  });
  autocode.action.toggleSection('imports');
  
  $('#imports-content-container .table a').remove();
  
  var import_index = 0, import_version;
  for (var import_name in autocode.object.sort(autocode.project.imports)) {
    import_version = autocode.project.imports[import_name];
    
    $('#imports-content-container .table').append(
      '<a class="file' + (autocode.data.current.import == import_name ? ' selected' : '') + '" onclick="autocode.action.loadImport({ repo: \'' + import_name + '\', index: ' + import_index + ' })">'
        + '<span class="image"><span class="icon" style="background-image: url(https://cdn.rawgit.com/' + import_name + '/master/.autocode/icon.svg)"></span></span>'
        + '<span class="info">'
          + '<span class="name">' + import_name + '</span>'
          + '<span class="generator">' + import_version + '</span>'
        + '</span>'
      + '</a>'
    );
    
    import_index++;
  }
  
  autocode.initState();
  
  if (autocode.project.imports) {
    $('#imports-content-container .content-right').show();
  } else {
    $('#imports-content-container .content-right').hide();
  }
  
  $('#imports-content-container').show();
  
  if (opts.disableSelected !== true && $(window).width() > autocode.mobile.minWidth) {
    $('#imports-content-container .table a').first().click();
  }
  
  if (!autocode.data.modules) {
    autocode.api.modules.get({
      success: function(modules) {
        autocode.data.modules = modules;
        
        $('#imports-search').keyup(function(e) {
          if (e.keyCode == 13) {
            if (!$(this).val().length) {
              return false;
            }
            var text = $(this).val();
            if ($('#fuzzy a').first().find('.text').length) {
              text = $('#fuzzy a').first().find('.text').text();
            }
            autocode.action.addImport({ repo: text });
            $(this).val('');
            autocode.fuzzy.close();
            return false;
          }
          
          var value = $(this).val();
          
          var mod, modules = [];
          for (var module_i in autocode.data.modules) {
            module_name = autocode.data.modules[module_i].name;
            if (autocode.project.imports && autocode.project.imports[module_name]) {
              continue;
            }
            if (module_name.match(new RegExp(value, 'i'))) {
              modules.push({
                action: {
                  name: 'addImport',
                  data: {
                    repo: module_name
                  }
                },
                icon: 'https://cdn.rawgit.com/' + module_name + '/master/.autocode/icon.svg',
                text: module_name
              });
            }
          }
          autocode.fuzzy.open({
            rows: modules,
            target: $(this),
            value: value
          });
        });
      }
    });
  }
};;autocode.state['interfaces'] = function() {
  if (autocode.project.interfaces) {
    $('#interfaces-content-container .table a').slice(1).remove();
    
    var interface_data, interface_name;
    for (var interface_i in autocode.project.interfaces) {
      interface_data = autocode.project.interfaces[interface_i];
      interface_name = interface_data.name;
      
      if (typeof(interface_name) == 'string' && interface_name.substr(0, 1) == '$' && autocode.project[interface_name.substr(1)]) {
        interface_name = autocode.project[interface_name.substr(1)];
      }
      
      $('#interfaces-content-container .table').append(
        '<a class="file" id="interfaces-' + interface_i + '" onclick="autocode.action.loadInterface({ index: \'' + interface_i + '\' })">'
          + '<span class="image">'
            + '<span class="icon interface-icon"></span>'
          + '</span>'
          + '<span class="info">'
            + '<span class="name">' + interface_name + '</span>'
          + '</span>'
        + '</a>'
      );
    }
    
    $('#interfaces-init').hide();
    $('#interfaces-content-container').show();
    
    $('#interfaces-content-container .table a').eq(1).click();
    
    $('#interfaces-port .value').text(interface_data.port || ' [ Click to Add ] ');
    
  } else {
    $('#interfaces-init').show();
    $('#interfaces-content-container').hide();
  }
  
  autocode.initState()
  
  autocode.action.toggleSection('interfaces');
};;autocode.state['output'] = function() {
  autocode.action.toggleColumn('output-content', 1);
  autocode.action.toggleSection('output');
  
  if (autocode.data.output || (autocode.project.outputs && autocode.project.imports)) {
    $('#output-content-container .table').empty();
    
    if (!autocode.data.output) {
      autocode.action.generate();
    } else {
      $('#output-content-container .content-left').replaceWith('<aside class="column content-left"></aside>');
      
      $('#output-content-container .content-left').on("changed.jstree", function (e, data) {
        if (data.node.children && data.node.children.length) {
          return;
        }
        
        autocode.ws.io.emit('contents', {
          config: autocode.project,
          project: autocode.repo.split('/')[1],
          user: autocode.repo.split('/')[0],
          file: data.instance.get_path(data.node,'/')
        });
      });
      $('#output-content-container .content-left').on("ready.jstree", function (e, data) {
        $('#output-content-container .icon').each(function() {
          var filename = $(this).parent().text();
          var icon = autocode.file.icon(filename);
          $(this).addClass(icon + '-icon');
        });
        
        $('.jstree-anchor').first().click();
      });
      
      $('#output-content-container .content-left').on("before_open.jstree", function (e, data) {
        $('#output-content-container .icon').each(function() {
          var filename = $(this).parent().text();
          var icon = autocode.file.icon(filename);
          $(this).addClass(icon + '-icon');
        });
      });
      
      $('#output-content-container .content-left').jstree({
        core: {
          data: autocode.data.output.files,
          multiple: false,
          themes: {
            dots: false
          }
        }
      });
    }
    
    $('#output-init').hide();
    $('#output-content-container').show();
    
  } else {
    $('#output-init').show();
    $('#output-content-container').hide();
  }
  
  autocode.resize.all();
};;;autocode.state['outputs'] = function() {
  if (autocode.project.outputs) {
    $('#outputs-content-container .table a').slice(1).remove();
    
    var output, output_filename, output_icon;
    for (var output_i in autocode.project.outputs) {
      output = autocode.project.outputs[output_i];
      
      if (autocode.project.outputs instanceof Array) {
        output_name = output.filename;
      } else {
        output_name = output_i;
      }
      
      if (output.filename) {
        output_filename = output.filename;
      } else if (autocode.project.exports && autocode.project.exports[output.generator.split('.')[0]]) {
        output_filename = autocode.project.exports[output.generator.split('.')[0]].filename;
      } else if (autocode.imports[output.generator.split('.')[0]] && autocode.imports[output.generator.split('.')[0]].exports) {
        output_filename = autocode.imports[output.generator.split('.')[0]].exports[output.generator.split('.')[1]].filename;
      } else {
        output_filename = '[ Untitled ]';
      }
      
      if (output.generator.split('.').length > 1) {
        output_icon = '<span class="icon" style="background-image: url(https://cdn.rawgit.com/crystal/' + output.generator.split('.')[0] + '/master/.autocode/icon.svg)"></span>'
      } else {
        output_icon = '<span class="icon project-icon black"></span>';
      }
      
      $('#outputs-content-container .table').append(
        '<a class="file" id="outputs-' + output_i + '" onclick="autocode.action.loadOutput({ output: \'' + output_i + '\' })">'
          + '<span class="image">'
            + output_icon
          + '</span>'
          + '<span class="info">'
            + '<span class="name">' + output_filename + '</span>'
            + '<span class="generator">' + output.generator + '</span>'
          + '</span>'
        + '</a>'
      );
    }
    
    autocode.initState()
    
    $('#outputs-init').hide();
    $('#outputs-content-container').show();
    
    autocode.action.loadOutput({ output: autocode.data.current.output || autocode.object.firstKey(autocode.project.outputs) });
    
  } else {
    $('#outputs-init').show();
    $('#outputs-content-container').hide();
    
    autocode.resize.outputsInit();
  }
  
  autocode.action.toggleSection('outputs');
};;;autocode.state['outputs/input/add'] = function(opts) {
  var field = $('#fuzzy').data('target');
  field.val('$' + opts.name).blur();
  
  autocode.fuzzy.close();
};;autocode.state['outputs/input/blur'] = function() {
  var field = $(this);
  var field_name = field.attr('name');
  var output = autocode.project.outputs[autocode.data.current.output];
  var value = field.val();
  
  field.unbind('keyup', autocode.state['outputs/input/fuzzy']);
  
  if (!output.spec) {
    output.spec = {};
  }
  
  if (field_name.match(/\[[0-9]*\]/)) {
    if (!output.spec[field_name.split('[')[0]]) {
      output.spec[field_name.split('[')[0]] = {};
    }
    if (field_name.match(/\[key\]/)) {
      var field_key = $(this).val();
      var field_value = $(this).next().val();
    } else {
      var field_key = $(this).prev().val();
      var field_value = $(this).val();
    }
    
    output.spec[field_name.split('[')[0]][field_key] = field_value;
    
  } else {
    output.spec[field_name] = value;
  }
  
  if (value.substr(0, 1) == '$' && autocode.project[value.substr(1)]) {
    value = autocode.project[value.substr(1)];
    field.addClass('variable').val(value);
  }
};;autocode.state['outputs/input/focus'] = function() {
  var field = $(this);
  var field_name = field.attr('name');
  var value = field.val();
  
  if (
    autocode.project.outputs[autocode.data.current.output].spec
    &&
    autocode.project.outputs[autocode.data.current.output].spec[field_name]
    &&
    autocode.project.outputs[autocode.data.current.output].spec[field_name].substr(0, 1) == '$'
  ) {
    value = autocode.project.outputs[autocode.data.current.output].spec[field_name];
  }
  
  field
    .bind('keyup', autocode.state['outputs/input/fuzzy'])
    .removeClass('variable')
    .val(value);
};;autocode.state['outputs/input/fuzzy'] = function(e) {
  var field = $(this);
  var value = field.val();
  
  if (value.substr(0, 1) != '$') {
    autocode.fuzzy.close();
    return;
  }
  
  value = value.substr(1);
  
  var variable, variables = [];
  for (var variable_name in autocode.project) {
    variable = autocode.project[variable_name];
    if (variable_name.match(new RegExp(value, 'i'))) {
      variables.push({
        icon: 'variable-icon',
        state: 'outputs/input/add?name=' + variable_name,
        text: variable_name
      });
    }
  }
  autocode.fuzzy.open({
    rows: variables,
    target: $(this),
    value: value
  });
};;;;autocode.state['outputs/property/add/choose'] = function(opts) {
  $('#popup input[name="property"]').val(opts.name);
  
  var value;
  switch (autocode.data.generators[autocode.data.current.generator].schema.properties[opts.name].type) {
    case 'boolean': {
      value = autocode.element.radio.html({
        options: {
          true: 'True',
          false: 'False'
        },
        name: 'value'
      });
      break;
    }
    default: {
      value = autocode.element.input.html({
        name: 'value',
        placeholder: autocode.data.generators[autocode.data.current.generator].schema.properties[opts.name].default
          ? autocode.data.generators[autocode.data.current.generator].schema.properties[opts.name].default
          : null
      });
      break;
    }
  }
  
  if ($('#popup .radio').length) {
    $('#popup .radio').replaceWith(value);
  } else {
    $('#popup input[name="value"]').replaceWith(value);
  }
  
  autocode.fuzzy.close();
};;;autocode.state['overview'] = function() {
  autocode.action.toggleColumn('overview-content', 1);
  if ($(window).width() > autocode.mobile.minWidth && (!$('#overview-content').find('.subcontent.selected').length || $('#overview-content').find('.subcontent.selected').index() == 1)) {
    autocode.state['overview/general']();
  } else {
    autocode.action.toggleSection('overview');
  }
};;;autocode.state['overview/author'] = function() {
  autocode.action.toggleColumn('overview-content', 2);
  
  $('#overview-author-name .value').text(autocode.project.author.name || ' [ Click to Add ]');
  $('#overview-author-email .value').text(autocode.project.author.email || ' [ Click to Add ]');
  $('#overview-author-url .value').text(autocode.project.author.url || ' [ Click to Add ]');
  
  autocode.action.toggleSection('overview', 'author');
};;autocode.state['overview/author/save'] = function() {
  if (!autocode.project.author) {
    autocode.project.author = {};
  }
  
  autocode.project.author.name = $('#overview-author-content input[name="name"]').val();
  autocode.project.author.email = $('#overview-author-content input[name="email"]').val();
  autocode.project.author.url = $('#overview-author-content input[name="url"]').val();
  autocode.project.copyright = $('#overview-author-content input[name="copyright"]').val();
  
  autocode.state['project/save']();
};;;autocode.state['overview/general'] = function() {
  autocode.action.toggleColumn('overview-content', 2);
  autocode.action.toggleSection('overview', 'general');
  
  $('#overview-general-name .value').text(autocode.project.name || '[ Click to Add ]');
  $('#overview-general-version .value').text(autocode.project.version || '[ Click to Add ]');
  $('#overview-general-description .value').html(marked(autocode.project.description) || '[ Click to Add ]');
  $('#overview-general-url .value').html(
    autocode.project.url
    ? '<a href="' + autocode.project.url + '">' + autocode.project.url + '</a>'
    : '[ Click to Add ]'
  );
  $('#overview-general-copyright .value').text(autocode.project.copyright || ' [ Click to Add ]');
};;autocode.state['overview/general/save'] = function() {
  autocode.project.name = $('#overview-general-content input[name="name"]').val();
  autocode.project.description = $('#overview-general-content textarea[name="description"]').val();
  autocode.project.url = $('#overview-general-content input[name="url"]').val();
  
  autocode.state['project/save']();
};;autocode.state.push = function(state, opts) {
  if (!autocode.state[state]) {
    return;
  }
  
  autocode.state[state](opts);
  
  if (autocode.listener.listeners[state]) {
    for (var listener_name in autocode.listener.listeners[state]) {
      autocode.listener.listeners[state][listener_name](opts);
    }
  }
};;;autocode.state['scripts'] = function() {
  $('#scripts-content-container').show();
  
  autocode.action.toggleSection('scripts');
  autocode.action.loadScript({ name: 'build' });
};;autocode.state['scripts/build'] = function() {
  autocode.action.loadScript({ name: 'build' });
};;autocode.state['scripts/run'] = function() {
  autocode.action.loadScript({ name: 'run' });
};;autocode.state['scripts/stop'] = function() {
  autocode.action.loadScript({ name: 'stop' });
};;autocode.state['tour'] = function() {
  autocode.popover.close();
  
  autocode.hint.locked = true;
  
  var click_event = navigator.userAgent.match(/mobile/i) ? 'touchend' : 'click';
  var hint_offset_top = -10;
  
  var import_check;
  var imports_click = function() {
    $(this).unbind(click_event, imports_click);
    
    autocode.hint.open({
      target: $('#imports-search'),
      text: 'Select a module to import into your project. For example, try typing "bower".'
    });
    
    $('#imports-search').keyup(imports_init_click);
    
    import_check = setInterval(function() {
      if ($('#imports-content-readme button').length) {
        clearInterval(import_check);
        delete(import_check);
        
        module_click();
      }
    }, 1000);
    
    autocode.resize.hint();
    
    $('#imports-search').focus();
  };
  
  var overview_click = function() {
    $(this).unbind(click_event, overview_click);
    
    $('#tour-content').hide();
    
    autocode.hint.open({
      minTop: 0,
      offsetTop: -10,
      target: $('#imports-tab'),
      text: 'Once done, click this tab to import modules.'
    });
    
    $('#imports-tab').bind(click_event, imports_click);
    
    autocode.resize.hint();
  };
  $('#overview-tab').bind(click_event, overview_click);
  
  var imports_init_click = function() {
    $(this).unbind('keyup', imports_init_click);
    
    autocode.hint.close();
    
    setTimeout(function() {
      $('#popup a').bind(click_event, module_click);
      
      autocode.resize.hint();
    }, 1000);
  };
  
  var check_hint;
  var module_click = function() {
    autocode.listener.remove('import_listener');
    
    check_hint = setInterval(function() {
      if ($('#imports-content-readme button').length) {
        autocode.hint.open({
          scrollUp: 'Scroll back up to click the button',
          target: function() {
            return $('#imports-content-readme button');
          },
          text: 'Click here to add an example from this module to your Autocode configuration.'
        });
        
        $('#imports-content-readme button').bind('click', example_click);

        autocode.resize.hint();
      } else {
        autocode.hint.close({ animated: false });
      }
    }, 100);
  };
  
  var example_click = function() {
    clearInterval(check_hint);
    
    $(this).unbind('click', example_click);
    
    autocode.hint.open({
      offsetTop: hint_offset_top,
      target: $('#generate-icon'),
      text: 'Great! Now let\'s generate some code.',
      top: $('#main').outerHeight()
    });
    
    $('#generate-icon').bind(click_event, generate_click);
    
    autocode.resize.hint();
  };
  
  var generate_click = function() {
    $(this).unbind(click_event, generate_click);
    
    autocode.hint.open({
      target: $('#user'),
      text: 'Woohoo! You used Autocode to generate code! Now login with GitHub to signup :)',
      top: 52
    });
    
    setTimeout(function() {
      autocode.hint.close();
    }, 10000);
    
    autocode.resize.hint();
  };
  
  $('#target-icon .text').text('Demo Server');
  
  $('#welcome').fadeOut(function() {
    $('.app').fadeIn();
    
    $('header').fadeIn();
    
    $('.info-icon').hide();
    
    autocode.ws.ip = '52.9.49.123';
    autocode.ws.init();
    
    autocode.repo = 'tour/autocode-' + Math.random();
    autocode.imports = {};
    autocode.data.generators = {};
    
    autocode.project = {
      name: 'My App',
      version: '1.0.0',
      description: 'My amazing new app.',
      url: 'https://example.org',
      author: {
        name: 'My Name',
        email: 'my.email@example.org',
        url: 'http//example.org/my-name',
        copyright: '2015 My Co.'
      }
    };
    autocode.data.originalConfig = '';
    
    //autocode.state['overview']();
    
    $('#tour-content').fadeIn();
    
    autocode.hint.open({
      offsetTop: -10,
      minTop: 0,
      target: $('#overview-tab'),
      text: 'Click here to fill out information about your project.'
    });
    
    autocode.resize.hint();
  });
};